import numpy as np, pandas as pd
from ijazz.RegionalFitter import RegionalFitter
import tensorflow as tf
from typing import Union, List, Dict
from pathlib import Path
import yaml, json
import argparse, sys
from ijazz.dtypes import floatzz, uintzz
from copy import deepcopy
from .sas_utils import to_correction_lib, json_safe_load, parameters_to_json, parameters_from_json, create_pt_corrector, apply_corrlib_df, compute_syst_from_jsons
from .categorize import categorize

def ijazz_sas_cmd():
    """
    entry point for the ijazz_sas command
    """
   
    parser = argparse.ArgumentParser(description=f'IJazZ Scale and Smearing fit')
    parser.add_argument('config', type=str, help='yaml config file')
    args = parser.parse_args(sys.argv[1:])

    with open(args.config, 'r') as fcfg:
        config = yaml.safe_load(fcfg)
    
    ijazz_sas(config)

def ijazz_sas(config:dict=None, do_fit=True, save_corrlib=True):
    """
    Perform the scale and smearing fit, plot the results and save the results.
    The configuration file should contain the following sections:

    - file_dt: data file (parquet format)
    - file_mc: mc file (parquet format)
    - dir_results: directory where to save the results
    - fitter: configuration for the fitter
    - sas: configuration for the scale and smearing fit
    - minimizer: configuration for the minimizer
    - syst: configuration for the systematic computation
    - corrlib: add description and version number to the correctionlib file
    - cat_latex: dictionary to convert the category name to latex
    - scale_flat_syst: flat systematic for scale
    - smear_flat_syst: flat systematic for smearing
    - dset_name: name of the dataset eg ``2023``
    - cset_name: name of the correction set eg ``EtaR9``

    See an example `sas_config.yaml <https://gitlab.cern.ch/fcouderc/ijazz_2p0/-/blob/master/config/sas_config.yaml?ref_type=heads>`_
    
    Args:
        config (dict): configuration dictionary containing the parameters for the fit
        do_fit: if True, perform the fit
        save_corrlib: if True, save the correctionlib file
    """

    files_dt = [config['file_dt']] if np.isscalar(config['file_dt']) else config['file_dt']
    files_mc = [config['file_mc']] if np.isscalar(config['file_mc']) else config['file_mc']

    dt = pd.concat([pd.read_parquet(afile).assign(ifile=ifile) for ifile, afile in enumerate(files_dt)]).reset_index(drop=True)
    mc = pd.concat([pd.read_parquet(afile).assign(ifile=ifile) for ifile, afile in enumerate(files_mc)]).reset_index(drop=True)

    dir_results = Path(config['dir_results'])
    dir_results.mkdir(parents=True, exist_ok=True)

    dset_name = config.get('dset_name', 'DSET')
    cset_name = config.get('cset_name', 'CSET')

    # -- save the raw mass for systematic computation
    mll_name = config['fitter'].get('name_mll', 'mee')
    dt[f'{mll_name}_orig'] = dt[mll_name]
    mc[f'{mll_name}_orig'] = mc[mll_name]

    config['sas']['dump_json'] = dir_results / f'SAS{dset_name}_syst-Nominal.json'
    if do_fit:
        compute_sas(dt, mc, config)

    json_file_out = dir_results / f'SAS{dset_name}_syst-Nominal.json'
    nominal_json = str(json_file_out)
    fit_2g_json = None
    import ijazz.plotting as ijp
    resp_range = (0.98, 1.03)
    reso_range = (0., 0.05)
    cat_latex = config.get('cat_latex', None)
    figs, _ = ijp.plot_results_from_json(json_file_out, resp_range=resp_range, reso_range=reso_range, 
                                         cat_latex=cat_latex, leg_fontsize='small')
    for ifig, fig in enumerate(figs):
        print(f" --> Plot nominal : {json_file_out.with_suffix(f'.fig{ifig}.jpg')}")
        fig.savefig(json_file_out.with_suffix(f'.fig{ifig}.jpg'))

    # -- plot the extra parameters if using double gaussian
    if config['fitter'].get('double_gaussian', False):
        figs, _ = ijp.plot_results_from_json2G(json_file_out, mu_range=(0.79, 1.01), reso2_range=(0.0, 0.15), frac_range=(0.79, 1.01), 
                                         cat_latex=cat_latex, leg_fontsize='small')
        for ifig, fig in enumerate(figs):
            print(f" --> Plot nominal : {json_file_out.with_suffix(f'.extra.fig{ifig}.jpg')}")
            fig.savefig(json_file_out.with_suffix(f'.extra.fig{ifig}.jpg'))
    
    if config.get('syst', False):
        syst_jsons = []

        scale_flat_syst = config.get('scale_flat_syst', 0)
        smear_flat_syst = config.get('smear_flat_syst', 0)
        config_ref = deepcopy(config)
        for syst_name, syst_cfg  in config['syst'].items():
            print(f'---------------------------------------')
            print(f'----- syst {syst_name}')
            print(f'---------------------------------------')
            dt[mll_name] = dt[f'{mll_name}_orig']
            mc[mll_name] = mc[f'{mll_name}_orig']
            print(syst_cfg)
            config = deepcopy(config_ref)
            for sub_sec, sub_cfg in syst_cfg.items():
                for k, v in sub_cfg.items():
                    config[sub_sec][k] = v
            config['sas']['correct_data'] = False
            config['sas']['correct_mc'] = False
            config['sas']['dump_json'] = dir_results / f'SAS{dset_name}_syst-{syst_name}.json'
            if do_fit:
                compute_sas(dt, mc, config)
            sas_syst = json_safe_load(config['sas']['dump_json'])

            if config['fitter'].get('double_gaussian', False):
                fit_2g_json = str(dir_results / f'SAS{dset_name}_syst-{syst_name}.json')
            else:
                syst_jsons.append(str(dir_results / f'SAS{dset_name}_syst-{syst_name}.json'))

            
    
            if config['fitter'].get('double_gaussian', False):
                # plot the scales and smearings ratio 
                figs, _ = ijp.plot_results_from_json({'nominal': nominal_json, 'syst': str(dir_results / f'SAS{dset_name}_syst-{syst_name}.json')}, 
                                        resp_range=(0.99, 1.01), reso_range=(0.5, 1.5), cat_latex=cat_latex, jsons_mode='ratio', leg_fontsize='small')
                for ifig, fig in enumerate(figs):
                    print(f" --> Plot {syst_name} : {dir_results}/SAS{dset_name}_syst-{syst_name}.2G-params.fig{ifig}.jpg")
                    fig.savefig(dir_results / f'SAS{dset_name}_syst-{syst_name}.2G-params.fig{ifig}.jpg')

                # plot the extra parameters if using double gaussian
                figs, _ = ijp.plot_results_from_json2G(parameters_from_json(sas_syst), 
                                       mu_range=(0.79, 1.01), reso2_range=(0.0, 0.15), frac_range=(0.79, 1.01), cat_latex=cat_latex, leg_fontsize='small')
                for ifig, fig in enumerate(figs):
                    print(f" --> Plot {syst_name} : {dir_results}/SAS{dset_name}_syst-{syst_name}.2G-params-extra.fig{ifig}.jpg")
                    fig.savefig(dir_results / f'SAS{dset_name}_syst-{syst_name}.2G-params-extra.fig{ifig}.jpg')
            
            else:
                #-- plot ratio of syst/nominal
                figs, _ = ijp.plot_results_from_json({'nominal': nominal_json, 'syst': str(dir_results / f'SAS{dset_name}_syst-{syst_name}.json')}, 
                                        resp_range=(0.99, 1.01), reso_range=(0.5, 1.5), cat_latex=cat_latex, jsons_mode='ratio', leg_fontsize='small')
                for ifig, fig in enumerate(figs):
                    print(f" --> Plot {syst_name} : {dir_results}/SAS{dset_name}_syst-{syst_name}.fig{ifig}.jpg")
                    fig.savefig(dir_results / f'SAS{dset_name}_syst-{syst_name}.fig{ifig}.jpg')
                
        # -- re-init config to its original value
        config = deepcopy(config_ref)

        if fit_2g_json:
            # -- compute the systematic errors from the jsons
            nominalWithSyst, fit2GWithSyst = compute_syst_from_jsons(nominal_json, syst_jsons, fit_2g=fit_2g_json, scale_flat_syst=scale_flat_syst, smear_flat_syst=smear_flat_syst)
            
            # save 2G fit with syst
            json_file_out2G = dir_results / f'SAS{dset_name}_syst-Nominal2GWithSyst.json'
            if save_corrlib:
                parameters_to_json(fit2GWithSyst, json_file_out2G)
            # save 1G fit with syst
            json_file_out = dir_results / f'SAS{dset_name}_syst-NominalWithSyst.json'
            if save_corrlib:
                parameters_to_json(nominalWithSyst, json_file_out)
            
            #-- plot 2G result with adjusted syst
            figs, _ = ijp.plot_results_from_json({'nominal':parameters_from_json(json_file_out2G), 'syst':parameters_from_json(json_file_out2G)}, jsons_mode='ratio',
                                              resp_range=(0.99, 1.01), reso_range=(0.5, 1.5), cat_latex=cat_latex, leg_fontsize='small')
            for ifig, fig in enumerate(figs):
                print(f" --> Plot Nominal 2G with Syst : {json_file_out2G.with_suffix(f'.fig{ifig}.jpg')}")
                fig.savefig(json_file_out2G.with_suffix(f'.fig{ifig}.jpg'))

    
        

            #-- plot 2G syst contribution
            # recompute mean and std of the 2G fit
            fit_2g = parameters_from_json(fit2GWithSyst)
            fit_2g['resp'] = fit_2g['mean2G']
            fit_2g['reso'] = fit_2g['std2G']

            #-- plot ratio of syst/nominal
            figs, _ = ijp.plot_results_from_json({'nominal': nominal_json, 'syst': fit_2g}, 
                                       resp_range=(0.99, 1.01), reso_range=(0.5, 1.5), cat_latex=cat_latex, jsons_mode='ratio', leg_fontsize='small')
            for ifig, fig in enumerate(figs):
                print(f" --> Plot {syst_name} : {dir_results}/SAS{dset_name}_syst-{syst_name}.fig{ifig}.jpg")
                fig.savefig(dir_results / f'SAS{dset_name}_syst-{syst_name}.fig{ifig}.jpg')

            # plot the mean difference
            figs, _ = ijp.plot_results_from_json({r'$\langle r \rangle_{1G}$':json_file_out,r'$\langle r \rangle_{2G}$':fit_2g}, jsons_mode='compare',param_to_plot='resp',resp_range=[0.995,1.01],reso_range=[0.0,0.045], cat_latex=cat_latex, hide_legend=False)
            for ifig, fig in enumerate(figs):
                print(f" --> Plot 2G syst (mean diff) : {dir_results / f'SAS{dset_name}_syst-smearing_shape-mean_diff.fig{ifig}.jpg'}")
                fig.savefig(dir_results / f'SAS{dset_name}_syst-smearing_shape-mean_diff.fig{ifig}.jpg')

            # plot the std difference
            figs, ax = ijp.plot_results_from_json({r'$\sigma_{1G}$':json_file_out,r'$\sigma_{2G}$':fit_2g}, jsons_mode='compare',param_to_plot='reso',resp_range=[0.99,1.015],reso_range=[0.0,0.045], cat_latex=cat_latex, hide_legend=False)
            for ifig, fig in enumerate(figs):
                print(f" --> Plot 2G syst (std diff) : {dir_results / f'SAS{dset_name}_syst-smearing_shape-std_diff.fig{ifig}.jpg'}")
                fig.savefig(dir_results / f'SAS{dset_name}_syst-smearing_shape-std_diff.fig{ifig}.jpg')
        else:
            # -- compute the systematic errors from the jsons
            nominalWithSyst = compute_syst_from_jsons(nominal_json, syst_jsons, scale_flat_syst=scale_flat_syst, smear_flat_syst=smear_flat_syst)
            json_file_out = dir_results / f'SAS{dset_name}_syst-NominalWithSyst.json'
            parameters_to_json(nominalWithSyst, json_file_out)

         #-- plot nominal result with adjusted syst
        figs, _ = ijp.plot_results_from_json({'nominal':parameters_from_json(json_file_out), 'syst':parameters_from_json(json_file_out)}, jsons_mode='ratio',
                                              resp_range=(0.99, 1.01), reso_range=(0.5, 1.5), cat_latex=cat_latex, leg_fontsize='small')
        for ifig, fig in enumerate(figs):
            print(f" --> Plot Nominal with Syst : {json_file_out.with_suffix(f'.fig{ifig}.jpg')}")
            fig.savefig(json_file_out.with_suffix(f'.fig{ifig}.jpg'))

    if config.get('corrlib', False) and save_corrlib:
        corrlib_file, cset_vars = to_correction_lib(json_file_out, dir_results=dir_results, dset_name=dset_name, cset_name=cset_name, 
                                                    cset_description=config['corrlib'].get('cset_description', None), 
                                                    cset_version=config['corrlib'].get('cset_version', 1))
        if fit_2g_json:
            _, _ = to_correction_lib(json_file_out2G, dir_results=dir_results, dset_name=dset_name+'2G', cset_name=cset_name, 
                                                        cset_description=config['corrlib'].get('cset_description', "SaS")+" using double gaussian smearing", 
                                                        cset_version=config['corrlib'].get('cset_version', 1))
        
        if config['corrlib'].get('smoothing', False):
            create_pt_corrector(json_file_out, dir_results=dir_results, dset_name=dset_name, cset_name=cset_name, 
                                cset_description=config['corrlib'].get('cset_description', None), 
                                cset_version=config['corrlib'].get('cset_version', 1))  

        def save_to_parquet(is_mc):
            correct = config['sas'].get('correct_mc', False) if is_mc else config['sas'].get('correct_data', False)
            if not correct:
                return
            
            df = mc if is_mc else dt
            syst_name = 'smear' if is_mc else 'scale'
            corr_name = f'EGMSmearAndSyst_{cset_name}_{dset_name}' if is_mc else f'EGMScale_{cset_name}_{dset_name}'
            print(f'Apply SAS correction (is_mc = {is_mc})')
            apply_corrlib_df(df, cset_vars, corr_name, corrlib_file, syst_name, mll_name=mll_name)

            cat_name = config['fitter'].get('name_cat', 'cat')
            col_to_drop = [col for col in df.columns if col[:len(cat_name)]==cat_name] + ['ifile'] 
            for ifile, afile in enumerate(files_mc if is_mc else files_dt):
                print(afile)
                df_fileout = Path(afile).with_suffix(f".{cset_name}Corr.parquet")
                print(f'Saving corrected dataset to {df_fileout}')
                df.query("ifile == @ifile").drop(columns=col_to_drop).to_parquet(df_fileout, engine='auto')
        save_to_parquet(is_mc=True)
        save_to_parquet(is_mc=False)      



    

class IJazZSAS:
    def __init__(self, dt: pd.DataFrame, mc: pd.DataFrame, config: dict, categories: dict=None) -> None:
        """This class performs the scale and smearing fit for leptons in different categories.
        The input dataframe should have as columns contain the categorisation variables.
        The naming convention is "var1" (resp. "var2") for the value of the variable var for lepton 1 (resp. 2)
        WARNING: If pt is used for categorisation the naming convention is "pt", so please rename if necessary
        For categorisation with "pt" bining it is more reliable to do the fit based on a categorisation on the relative pt.
        This will be automatically done unless turned off (use_r_pt=False).

        Args:
            dt (pd.DataFrame): dataframe containing the data events
            mc (pd.DataFrame): dataframe containing the mc events
            config (dict): dictionary containing the config of the RegionalFitter and its minimize function. An example yaml is given in the repertory configs/ 
            categories (dict): dictionary for categorisation, e.g. {'pt': [25, 50, 100], 'abs_eta': [0, 1, 2]}. Defaults to None (can be specified in the config dict)

        NB: the config dictionnary can be saved under a yaml file (prefered), opened and passed to the fitter.
        It **MUST** have three section: "sas", "fitter", "minimizer"
        """
        try:
            config["sas"]
        except KeyError:
            raise KeyError('The sas section is not present in the configuration dict')
        
        try:
            config["fitter"]
        except KeyError:
            raise KeyError('The fitter section is not present in the configuration dict')

        try:
            config["minimizer"]
        except KeyError:
            raise KeyError('The minimizer section is not present in the configuration dict')

        if categories is None:
            # - this will crask if no categories section but this is expected.
            categories = config['sas']['categories']

        self.categories_def = categories
        self.cfg = config
        self.double_gaussian = config['fitter'].get('double_gaussian', False)
        print(f"Double gaussian fit: {self.double_gaussian}")

        use_rpt = config['sas'].get('use_rpt', True)
        self.mll_name = config['fitter'].get('name_mll', 'mee')
        self.pt_name = config['sas'].get('name_pt_var', 'pt')
        self.cat_name = config['fitter'].get('name_cat', 'cat')
        cut = config['sas'].get('cut', None)
        cut = '' if cut is None else cut

        if categories.get(self.pt_name, None):
            pt_bins = categories[self.pt_name]
            cut_pt = f"{self.pt_name}1 >= {pt_bins[0]} and {self.pt_name}2 >= {pt_bins[0]}" + " and " + \
                f"{self.pt_name}1 < {pt_bins[-1]} and {self.pt_name}2 < {pt_bins[-1]}"
            cut = cut_pt if cut == "" else f'{cut} and {cut_pt}'
        self.cut = cut
        print(f"Applying selection: {cut}")

        # -- if one of the categories is pt or et, categorize first vs pT/mee
        self.has_pt = False
        if use_rpt:
            self.categories = {'r_pt'  if key == self.pt_name else key : np.array(values)/91.2 
                                if key == self.pt_name else np.array(values) for key, values in categories.items()}
            if self.pt_name in list(categories.keys()):
                self.has_pt = True
                for df in [dt, mc]:
                    for i_ele in [1, 2]:
                        df[f'r_pt{i_ele}'] = df[f'{self.pt_name}{i_ele}'] / df[self.mll_name]
        else:
            self.categories = categories
        
        # can have different MC/data categories when recasting w/r to pT
        self.categories_mc = self.categories.copy() 
        self.categories_dt = self.categories.copy() 

        for df in [dt, mc]:
            self.lepton_reg = categorize(df, self.categories, cut=self.cut, prefix=self.cat_name)

        self.dt = dt
        self.mc = mc

        self.fitter = None
        self.resp = None
        self.reso = None
        self.resp_corr = None     # -- response after data scale correction
        self.reso_uncorr = None   # -- resolution before data scale correction
        self.eresp = None
        self.ereso = None
        self.eresp_mc = None
        self.ereso_mc = None
        self.pt_mean = None

    def fit(self, optimizer=tf.keras.optimizers.Adam) -> np.ndarray:
        """Actually performs the fit

        Args:
            optimizer (class): tensorflow compatible optimizer for instance tf.keras.optimizers.Adam (this is not the optimizer the the class)
            learning_rate (float, optional): learning rate to pass to the optimizer. Defaults to 0.01.
            dnll_tol (float, optional): delta likelihood tolerance to stop the fit. Defaults to 0.01.
            max_epochs (int, optional): maximum number of iteration. Defaults to 1000.
            init_rand (bool, optional): random starting poin. Defaults to False.
            batch_size (int, optional): can perform the fit in batch size (usefull for large number of parameters). Defaults to -1.
            device (str, optional): _description_. Defaults to 'CPU'.
            batch_training (bool, optional): use the batch training mode. Defaults to False.
            hess (str, optional): computation of the hession (None, 'numerical', or 'analytical'), it can take a while. Defaults to 'numerical'.
            cats (_type_, optional): subset of categories to perform the fit. Defaults to slice(None).

        Returns:
            np.ndarray: array of the value of the -2 * log likelihood during the fit
        """
        learning_rate = self.cfg['sas'].get('learning_rate', 1e-3)
        print("======== FIT (first fit, binings including pt require a second one)")
        fitter = RegionalFitter(self.dt, self.mc, n_par=self.lepton_reg.size, **self.cfg['fitter'])
        nlls = fitter.minimize(optimizer(learning_rate=learning_rate), **self.cfg['minimizer'])       
        self.resp = fitter.resp.numpy().astype(np.float32).copy()
        self.reso = np.abs(fitter.reso.numpy().astype(np.float32).copy())
        self.eresp = np.zeros(self.resp.shape, dtype=np.float32)
        self.ereso = np.zeros(self.reso.shape, dtype=np.float32)
        self.eresp_mc = np.ones(self.resp.shape, dtype=np.float32)*np.nan
        self.ereso_mc = np.ones(self.reso.shape, dtype=np.float32)*np.nan

        if self.double_gaussian:
            self.reso2 = np.abs(fitter.reso2.numpy().astype(np.float32).copy())
            self.mu = fitter.mu.numpy().astype(np.float32).copy()
            self.frac = fitter.frac.numpy().astype(np.float32).copy()
            self.ereso2 = np.zeros(self.reso2.shape, dtype=np.float32)
            self.emu = np.zeros(self.mu.shape, dtype=np.float32)
            self.efrac = np.zeros(self.frac.shape, dtype=np.float32)
            self.ereso2_mc = np.ones(self.reso2.shape, dtype=np.float32)*np.nan
            self.emu_mc = np.ones(self.mu.shape, dtype=np.float32)*np.nan
            self.efrac_mc = np.ones(self.frac.shape, dtype=np.float32)*np.nan

        if self.has_pt:
            print("======== FIT with scale correction for pT bining")
            self.recast_pt_bins(prefix='cat_corr')
            self.correct_data(self.resp, prefix='cat_corr')
            fitter = RegionalFitter(self.dt, self.mc, n_par=self.lepton_reg.size, **self.cfg['fitter'])
            nlls = fitter.minimize(optimizer(learning_rate=learning_rate), **self.cfg['minimizer'])
            self.reso_uncorr = self.reso.copy()
            self.resp_uncorr = self.resp.copy()
            self.resp = self.resp * fitter.resp.numpy()
            self.reso = np.abs(fitter.reso.numpy().copy())

            if self.double_gaussian:
                self.reso2_uncorr = self.reso2.copy()
                self.mu_uncorr = self.mu.copy()
                self.frac_uncorr = self.frac.copy()
                self.reso2 = np.abs(fitter.reso2.numpy().copy())
                self.mu = fitter.mu.numpy().copy()
                self.frac = fitter.frac.numpy().copy()

            # -- uncorrect the data
            self.dt[self.mll_name] = self.dt[f'{self.mll_name}_raw']
            for i_ele in [1, 2]:
                self.dt[f'{self.pt_name}{i_ele}'] = self.dt[f'pt_raw{i_ele}']
            self.dt.drop(columns=[f'{self.mll_name}_raw', 'pt_raw1', 'pt_raw2'], inplace=True)

        self.fitter = fitter  # -- keep the fitter
        return nlls

    def calc_err_stat(self, force=False):
        hess = self.cfg['sas'].get('hess', False)
        if hess:
            numerical = True if hess == 'numerical' else False
            print(f"======== HESSIAN COMPUTATION: numerical {numerical}")
            self.fitter.covariance(numerical=numerical, force=force, batch_size=-1)
            self.eresp = self.fitter.eresp
            self.ereso = self.fitter.ereso
            if self.double_gaussian:
                self.ereso2 = self.fitter.ereso2
                self.emu = self.fitter.emu
                self.efrac = self.fitter.efrac
        else:
            print("NB: HESSIAN computation was not requested")

    def calc_err_mc(self):
        """compute the error due to MC statistical fluctuations, this may take a while
        """
        if self.cfg['sas'].get('err_mc', True):
            print(f"======== MC UNCERTAINTY COMPUTATION")
            self.fitter.calc_err_mc()
            self.eresp_mc = self.fitter.eresp_mc.copy()
            self.ereso_mc = self.fitter.ereso_mc.copy()
            self.eresp = np.sqrt(self.eresp**2 + self.eresp_mc**2)
            self.ereso = np.sqrt(self.ereso**2 + self.ereso_mc**2)
            if self.double_gaussian:
                self.ereso2_mc = self.fitter.ereso2_mc.copy()
                self.emu_mc = self.fitter.emu_mc.copy()
                self.efrac_mc = self.fitter.efrac_mc.copy()
                self.ereso2 = np.sqrt(self.ereso2**2 + self.ereso2_mc**2)
                self.emu = np.sqrt(self.emu**2 + self.emu_mc**2)
                self.efrac = np.sqrt(self.efrac**2 + self.efrac_mc**2)
        else:
            print("NB: no MC uncertainties were required")

    def correct_data(self, response: np.ndarray, prefix:str=None):
        """correct the data dataframe after the fit

        Args:
            response (np.ndarray): array with the responses to apply for correction
            prefix (str, optional): prefix used for naming the categorisation. Defaults to None.
        """
        if prefix is None:
            prefix = self.cfg['fitter']['name_cat']
        self.dt[f'{self.mll_name}_raw'] = self.dt[self.mll_name]
        for i_ele in [1, 2]:
            self.dt[f'pt_raw{i_ele}'] = self.dt[f'{self.pt_name}{i_ele}']
            idx = self.dt[f'{prefix}{i_ele}'] >= 0
            scale = response[self.dt.loc[idx, f'{prefix}{i_ele}']]
            self.dt.loc[idx, f'{self.pt_name}{i_ele}'] /= scale
            self.dt.loc[idx, self.mll_name] /= np.sqrt(scale)

    def correct_mc(self, resolution: np.ndarray, prefix:str=None):
        """correct the mc dataframe

        Args:
            resolution (np.ndarray): array with the relative resolution to smear the MC
            prefix (str, optional): prefix used for naming the categorisation. Defaults to None.
        """
        if prefix is None:
            prefix = self.cfg['fitter']['name_cat']

        

        def rbinorm(n, mu, sigma1, sigma2, frac):
            # Generate samples from each normal distribution
            scales = np.array([mu*np.random.normal(1, sigma2, n), np.random.normal(1, sigma1, n)])
            binom = np.random.binomial(1, frac, n)
            return scales[binom, np.arange(n)]

        self.mc[f'{self.mll_name}_raw'] = self.mc[self.mll_name]
        for i_ele in [1, 2]:
            self.mc[f'pt_raw{i_ele}'] = self.mc[f'{self.pt_name}{i_ele}']
            idx = self.mc[f'{prefix}{i_ele}'] >= 0
            if self.cfg['fitter'].get('double_gaussian', False):
                # smear = rbinorm(len(idx), self.mu[self.mc.loc[idx, f'{prefix}{i_ele}']], resolution[self.mc.loc[idx, f'{prefix}{i_ele}']], self.reso2[self.mc.loc[idx, f'{prefix}{i_ele}']], self.frac[self.mc.loc[idx, f'{prefix}{i_ele}']])
                smear = rbinorm(len(idx), self.mu[0], resolution[self.mc.loc[idx, f'{prefix}{i_ele}']], self.reso2[self.mc.loc[idx, f'{prefix}{i_ele}']], self.frac[0])
            else:
                smear = np.random.normal(1, resolution[self.mc.loc[idx, f'{prefix}{i_ele}']]).astype(np.float32)
            self.mc.loc[idx, f'{self.pt_name}{i_ele}'] *= smear
            self.mc.loc[idx, self.mll_name] *= np.sqrt(smear)
    
    def recast_pt_bins(self, prefix='cat'):
        """when doing a fit that contains pt bining with the option use_r_pt, one has to recast the relative pt in real pt value (in GeV)

        Args:
            prefix (str, optional): prefix used for naming the categorisation. Defaults to 'cat'.
        """
        if not self.has_pt:
            return 

        prefix_ref = self.cfg['fitter'].get('name_cat', 'cat')

        pt_bins = []
        for idf, df in enumerate([self.dt, self.mc]):
            df_cat_pt = pd.concat( [df[[f'{prefix_ref}_r_pt{ilep}', f'{self.pt_name}{ilep}' , f'{prefix_ref}{ilep}']]\
                .rename(columns={f'{prefix_ref}_r_pt{ilep}': f'{prefix_ref}_r_pt', 
                                 f'{self.pt_name}{ilep}': self.pt_name, 
                                 f'{prefix_ref}{ilep}': prefix_ref}) for ilep in [1, 2]] )
            pt_mean = df_cat_pt.query('cat >= 0').groupby(f'{prefix_ref}_r_pt')[self.pt_name].mean().to_numpy()
            pt_bins_def = self.categories_def[self.pt_name]
            pt_bins = np.concatenate([[pt_bins_def[0]], (pt_mean[:-1]+pt_mean[1:])*0.5, [pt_bins_def[-1]]])
            categories = {key: pt_bins if key == self.pt_name else np.array(values) for key, values in self.categories_def.items()}
            if idf == 0:
                self.categories_dt = categories.copy()
            else:
                self.categories_mc = categories.copy()
            self.lepton_reg = categorize(df, categories, prefix=prefix, cut=self.cut)
            if idf == 0:
                self.pt_mean = pt_mean

    def dump_json(self, outfile: Union[Path, str]=None):
        """Dump the resulst to a json file

        Args:
            outfile (Union[Path, str], optional): output file name. Defaults to None.
        """
        if outfile is None:
            return
        categories = self.categories_dt if self.has_pt else self.categories
        if self.double_gaussian:
            def reshape(param):
                if param.shape[0] == 1:
                    return np.ones((self.fitter.n_par,), dtype=np.float64)*param
                else:
                    return param
            parameters_to_json({"categories": categories, "resp": self.resp, "reso": self.reso, "reso2": reshape(self.reso2), "mu": reshape(self.mu), "frac": reshape(self.frac),
                                "eresp": self.eresp, "ereso": self.ereso, "ereso2": reshape(self.ereso2), "emu": reshape(self.emu), "efrac": reshape(self.efrac),
                                "eresp_mc": self.eresp_mc, "ereso_mc": self.ereso_mc, "ereso2_mc": reshape(self.ereso2_mc), "emu_mc": reshape(self.emu_mc), "efrac_mc": reshape(self.efrac_mc), "pt_mean": self.pt_mean}, outfile)
        else:
            parameters_to_json({"categories": categories, "resp": self.resp, "reso": self.reso,
                                "eresp": self.eresp, "ereso": self.ereso,
                                "eresp_mc": self.eresp_mc, "ereso_mc": self.ereso_mc, "pt_mean": self.pt_mean}, outfile)


def compute_sas(dt, mc, cfg, categories=None) -> IJazZSAS:
    """Wrapper to compute the scale and smearing parameters

    Args:
        dt (pd.DataFrame): input dataframe for data
        mc (pd.DataFrame): input dataframe for mc
        cfg (Union[str, Path]): config file to control the sas fit
        categories (dict, optional): categories to do the fit (can be specified in config). Defaults to None.

    Returns:
        IJazZSAS: return the ScaleAndSmearing fitter
    """
    sas = IJazZSAS(dt, mc, config=cfg, categories=categories)
    sas.fit()
    sas.calc_err_stat()
    sas.calc_err_mc()
    sas.dump_json(cfg['sas'].get('dump_json', None))
    # if cfg['sas'].get('correct_data', False):
    #     sas.recast_pt_bins()
    #     sas.correct_data(sas.resp)
    # if cfg['sas'].get('correct_mc', False):
    #     sas.correct_mc(sas.reso)
    return sas


class IJazZEnergyCorrector:
    def __init__(self, dt: pd.DataFrame, mc: pd.DataFrame, json_file: Union[str, Path, Dict], mll_name, pt_name) -> None:
        if not isinstance(json_file, Dict):
            with open(json_file) as f_json:
                json_file = json.load(f_json)

        self.categories = json_file['categories']
        self.resp = np.array(json_file['resp'], dtype=np.float32).flatten()
        self.reso = np.array(json_file['reso'], dtype=np.float32).flatten()
        self.eresp = np.array(json_file['eresp'], dtype=np.float32).flatten()
        self.ereso = np.array(json_file['ereso'], dtype=np.float32).flatten()

        self.mll_name = mll_name
        self.pt_name = pt_name
        self.prefix = "LepCatCorr"
        for df in [dt, mc]:
            self.lepton_cat = categorize(df, self.categories, prefix=self.prefix)
        self.mc = mc
        self.dt = dt

    def reset_data(self):
        self.dt[self.mll_name] = self.dt[f'{self.mll_name}_raw']

    def reset_mc(self):
        self.mc[self.mll_name] = self.mc[f'{self.mll_name}_raw']

    def correct_data(self):
        self.dt[f'{self.mll_name}_raw'] = self.dt[self.mll_name]
        for i_ele in [1, 2]:
            self.dt[f'pt_raw{i_ele}'] = self.dt[f'{self.pt_name}{i_ele}']
            idx = self.dt[f'{self.prefix}{i_ele}'] >= 0
            scale = self.resp[self.dt.loc[idx, f'{self.prefix}{i_ele}']]
            self.dt.loc[idx, f'{self.pt_name}{i_ele}'] /= scale
            self.dt.loc[idx, self.mll_name] /= np.sqrt(scale)
    
    def correct_mc(self, do_syst=False):
        self.mc[f'{self.mll_name}_raw'] = self.mc[self.mll_name]
        for i_ele in [1, 2]:
            self.mc[f'pt_raw{i_ele}'] = self.mc[f'{self.pt_name}{i_ele}']
            idx = self.mc[f'{self.prefix}{i_ele}'] >= 0
            smear = np.random.normal(1, self.reso[self.mc.loc[idx, f'{self.prefix}{i_ele}']]).astype(np.float32)
            self.mc.loc[idx, f'{self.pt_name}{i_ele}'] *= smear
            self.mc.loc[idx, self.mll_name] *= np.sqrt(smear)

        if do_syst:
            self.mc[f'{self.mll_name}_scaleUp'] = self.mc[self.mll_name]
            self.mc[f'{self.mll_name}_scaleDo'] = self.mc[self.mll_name]
            for i_ele in [1, 2]:
                idx = self.mc[f'{self.prefix}{i_ele}'] >= 0
                eresp = self.eresp[self.mc.loc[idx, f'{self.prefix}{i_ele}']]
                self.mc.loc[idx, f'{self.mll_name}_scaleUp'] /= np.sqrt(1. + eresp)
                self.mc.loc[idx, f'{self.mll_name}_scaleDo'] /= np.sqrt(1. - eresp)

            self.mc[f'{self.mll_name}_smearUp'] = self.mc[f'{self.mll_name}_raw']
            self.mc[f'{self.mll_name}_smearDo'] = self.mc[f'{self.mll_name}_raw']
            for i_ele in [1, 2]:
                idx = self.mc[f'{self.prefix}{i_ele}'] >= 0
                reso = self.reso[self.mc.loc[idx, f'{self.prefix}{i_ele}']]
                ereso = self.ereso[self.mc.loc[idx, f'{self.prefix}{i_ele}']]
                smear_up = np.random.normal(1, (reso+ereso)[self.mc.loc[idx, f'{self.prefix}{i_ele}']]).astype(np.float32)
                smear_do = np.random.normal(1, (reso-ereso)[self.mc.loc[idx, f'{self.prefix}{i_ele}']]).astype(np.float32)

                self.mc.loc[idx, f'{self.mll_name}_smearUp'] *= np.sqrt(smear_up)
                self.mc.loc[idx, f'{self.mll_name}_smearDo'] *= np.sqrt(smear_do)



import argparse
import sys
import yaml
import numpy as np
from pathlib import Path
from copy import deepcopy



def merge_dicts(dict1, dict2):
    """Recursively merges dict2 into dict1."""
    for key, value in dict2.items():
        if isinstance(value, dict) and key in dict1:
            merge_dicts(dict1[key], value)
        else:
            if key == 'cut' and dict2.get('use_year_selection', True):
                dict1[key] = dict1.get(key, '') + ' and ' + value
            elif key =='use_year_selection':
                pass
            else:
                dict1[key] = value

def ijazz_config_sas():
    """Entry point for the IJazZ Scale and Smearing fit configuration script.
    This script reads the provided configuration file and generates YAML files for each step.
    """
    parser = argparse.ArgumentParser(description=f'IJazZ Scale and Smearing fit')
    parser.add_argument('config', type=str, help='yaml config file')
    parser.add_argument('--cfg', type=str, default=None, help='path to the yaml config with steps')
    args = parser.parse_args(sys.argv[1:])

    with open(args.config, 'r') as fcfg:
        config = yaml.safe_load(fcfg)

    with open(args.cfg, 'r') as fcfg:
        cfg = yaml.safe_load(fcfg)
        
    config_sas(config, cfg)

def config_sas(config: dict,cfg: dict):
    """Creates YAML configuration files for SAS steps 
    based on the provided `config` and `cfg` dictionaries. This function processes 
    datasets, applies cuts, and generates YAML files for each step in the configuration.
    
    Args:
        config (dict): A dictionary containing the main configuration. 
            Expected keys include:
            - 'datasets': List of dataset dictionaries with 'file_dt' and 'file_mc' keys.
            - 'dir_yaml': Directory path to save the generated YAML files.
            - 'sas': Dictionary containing SAS-specific configurations (e.g., 'cut').
            - 'object_type': A string representing the object type.
            - 'year': A string representing the year.
            - 'dir_results': Directory path for storing results.
        cfg (dict): A dictionary containing the SAS steps configuration. 
            Expected keys include:
            - 'steps': List of step dictionaries, each containing:
                - 'name': Name of the step.
                - 'split': Boolean indicating whether to split datasets.
                - 'sas': Dictionary with SAS-specific step configurations 
                    (e.g., 'correct_data').
    Returns:
        None: The function writes YAML files to the specified directory.
        
    Example:
        config = {
            'datasets': [{'file_dt': ['data1.parquet'], 'file_mc': ['mc1.parquet']}],
            'dir_yaml': './yaml_configs',
            'sas': {'cut': 'some_cut'},
            'object_type': 'Pho',
            'year': '2023',
            'dir_results': './results'
        }
        cfg = {
            'steps': [{'name': 'Step1', 'split': False, 'sas': {'correct_data': True}}]
        }
        config_sas(config, cfg)
    """
    datasets_sas = config['datasets']

    dir_yaml = Path(config.get('dir_yaml','.'))
    dir_yaml.mkdir(parents=True, exist_ok=True)

    sas = config.get('sas', None)
    corr_name = ''

    if config and (cut := sas.get('cut', None)):
        print(f'Applying cut: {cut}')
        cfg['sas']['cut'] = cut

    for i,step in enumerate(cfg['steps']):
        print(f'Processing step: {step["name"]}')
        # print(step['sas'].get('correct_data',True))
        split = step.get('split', False)
        for dataset in datasets_sas:
            dataset['file_dt'] = [file.replace('.parquet', f'.{corr_name}.parquet' if (i and cfg['steps'][max(0,i-1)]['sas'].get('correct_data',True)) else '.parquet') for file in dataset['file_dt']]
            
        if split:
            print('Used split datasets')
            datasets = datasets_sas
            
        else:
            files_dt = []
            files_mc = []
            for dataset in datasets_sas:
                files_dt += [dataset['file_dt']] if np.isscalar(dataset['file_dt']) else dataset['file_dt']
                files_mc += [dataset['file_mc']] if np.isscalar(dataset['file_mc']) else dataset['file_mc']
            
            
            datasets = [{"subyear": '','file_dt': files_dt, 'file_mc': files_mc}]
        
        for dataset in datasets:
            file_dt = [dataset['file_dt']] if np.isscalar(dataset['file_dt']) else dataset['file_dt']
            file_mc = [dataset['file_mc']] if np.isscalar(dataset['file_mc']) else dataset['file_mc']
    
            corr_name = config['object_type'] + step['name'] + 'Corr'
            cset_name = config['object_type'] + step['name']
            dset_name = config['year'] + dataset.get('subyear','')
            
            dir_results = Path(config['dir_results']) / step['name']
            

            config_step = deepcopy(cfg)
            merge_dicts(config_step, step)
            config_step['file_dt'] = file_dt
            config_step['file_mc'] = file_mc
            config_step['dir_results'] = str(dir_results)
            config_step['dset_name'] = dset_name
            config_step['cset_name'] = cset_name
            config_step.pop('steps', None)
            # print(config_step)

            
            name_yaml = f'sas_{step["name"]}{"_" if dataset.get("subyear","") else ""}{dataset.get("subyear","")}.yaml'
            print(f'Writing yaml file: {dir_yaml / name_yaml}')
            with open(dir_yaml / name_yaml, 'w') as yaml_file:
                yaml.dump(config_step, yaml_file, default_flow_style=False, sort_keys=False)


import numpy as np
import tensorflow as tf
from pathlib import Path
import json, yaml
import pandas as pd
from typing import Dict, Union, List, Tuple
import matplotlib.pyplot as plt
import cms_fstyle as cms
from cms_fstyle.fitter import FFitter
from copy import deepcopy

def ijazz_shape(categories: dict) -> List:
    """Returns the shape to be used for reshaping the tensor of parameters of the ijazz fit

    Args:
        categories (dict): categories used in the ijazz sas fit

    Returns:
        list: shape
    """
    return [len(bins)-1 for bins in categories.values()]

    
def parameters_from_json(json_file: Union[str, Path, Dict]) -> Dict:
    """Extract and re-shape the IJazZ sas results saved in a json file

    Args:
        json_file (Union[str, Path, Dict]): name of the json file where the fit results are stored

    Returns:
        dict: dictionnary with the categories used in the fit and the results of the fit
    """

    if isinstance(json_file, Dict):
        ijazz_json = json_file.copy()
    else:
        with open(json_file) as f_json:
            ijazz_json = json.load(f_json)

    categories = ijazz_json['categories']
    shape = ijazz_shape(ijazz_json['categories'])

    
    par_names = ['resp', 'reso', 'eresp', 'ereso', 'eresp_mc', 'ereso_mc'] \
                +(['reso2', 'ereso2','ereso2_mc', 'mu', 'emu', 'emu_mc', 'frac', 'efrac', 'efrac_mc'] if 'reso2' in ijazz_json.keys() else []) \
                +(['mean2G', 'std2G'] if 'mean2G' in ijazz_json.keys() else [])
    parameters = {par_name: np.array(ijazz_json.get(par_name, np.ones(shape)*np.nan)).reshape(shape) for par_name in par_names}
    parameters['eresp'] = np.where(np.isnan(parameters['eresp']), np.inf, parameters['eresp'])
    parameters['ereso'] = np.where(np.isnan(parameters['ereso']), np.inf, parameters['ereso'])
    parameters = {k: tf.convert_to_tensor(v, dtype=tf.float32) for k, v in parameters.items()}

    parameters['bins'] = [np.array(bin) for bin in categories.values()]
    parameters['bins'] = [np.where(bin <= -9999, -np.inf, bin) for bin in parameters['bins']]
    parameters['bins'] = [np.where(bin >= +9999, +np.inf, bin) for bin in parameters['bins']]
    parameters['bins'] = [tf.convert_to_tensor(bin, dtype=tf.float32) for bin in parameters['bins'] ]

    parameters['dims'] = np.array(list(categories.keys()), dtype=str)
    parameters['pt'] = np.array(ijazz_json.get('pt_mean', None))

    return parameters


def parameters_to_json(params:Dict, outfile: Union[str, Path]) -> None:
    """Dump the parameters dictionnary to a json file

    Args:
        params (Dict): input dictionnary
        outfile (Union[str, Path]): output file
    """
    print(f'Dump results to json file: {outfile}')
    categories = params['categories']
    categories_serialized = {}
    for k, v in categories.items():
        v = np.where(np.array(v) == +np.inf, +9999, np.array(v))
        v = np.where(np.array(v) == -np.inf, -9999, np.array(v))
        categories_serialized[k] = v.tolist()
    categories_serialized

    with open(outfile, "w")  as fjson:
        eresp = np.where(np.isnan(np.array(params['eresp'])), 0, np.array(params['eresp']))
        ereso = np.where(np.isnan(np.array(params['ereso'])), 0, np.array(params['ereso']))
        eresp_mc = np.where(np.isnan(np.array(params['eresp_mc'])), 0, np.array(params['eresp_mc']))
        ereso_mc = np.where(np.isnan(np.array(params['ereso_mc'])), 0, np.array(params['ereso_mc']))

        pt_mean = None if params.get('pt_mean', None) is None else params['pt_mean'].tolist()

        if "reso2" in params.keys():
            emu = np.where(np.isnan(np.array(params['emu'])), 1, np.array(params['emu']))
            ereso2 = np.where(np.isnan(np.array(params['ereso2'])), 1, np.array(params['ereso2']))
            efrac = np.where(np.isnan(np.array(params['efrac'])), 1, np.array(params['efrac']))
            emu_mc = np.where(np.isnan(np.array(params['emu_mc'])), 1, np.array(params['emu_mc']))
            ereso2_mc = np.where(np.isnan(np.array(params['ereso2_mc'])), 1, np.array(params['ereso2_mc']))
            efrac_mc = np.where(np.isnan(np.array(params['efrac_mc'])), 1, np.array(params['efrac_mc']))
            json.dump({"categories": categories_serialized,
                "resp": np.array(params['resp']).tolist(), "reso": np.array(params['reso']).tolist(),
                "reso2": np.array(params['reso2']).tolist(), "mu": np.array(params['mu']).tolist(), "frac" : np.array(params['frac']).tolist(),
                "eresp": eresp.tolist(), "ereso": ereso.tolist(),
                "ereso2": ereso2.tolist(), "emu": emu.tolist(), "efrac" : efrac.tolist(),
                "eresp_mc": eresp_mc.tolist(), "ereso_mc": ereso_mc.tolist(),
                "ereso2_mc": ereso2_mc.tolist(), "emu_mc": emu_mc.tolist(), "efrac_mc" : efrac_mc.tolist(),
                "pt_mean": pt_mean} | ({"mean2G": np.array(params['mean2G']).tolist(), "std2G": np.array(params['std2G']).tolist()} if "mean2G" in params.keys() else {}),
                fjson, indent=4)
        else:
            json.dump({"categories": categories_serialized,
                "resp": np.array(params['resp']).tolist(), "reso": np.array(params['reso']).tolist(),
                "eresp": eresp.tolist(), "ereso": ereso.tolist(),
                "eresp_mc": eresp_mc.tolist(), "ereso_mc": ereso_mc.tolist(),
                "pt_mean": pt_mean},
                fjson, indent=4)


def json_safe_load(json_name, mc_errors=False):
    with open(json_name, 'r') as fjson:
        sas = json.load(fjson)
    for k in ['resp', 'reso', 'eresp', 'ereso', 'pt_mean'] + \
            (['eresp_mc', 'ereso_mc'] if mc_errors else []) + \
            (['reso2', 'ereso2', 'mu', 'emu', 'frac', 'efrac'] if 'reso2' in sas.keys() else []) + \
            (['ereso2_mc', 'emu_mc', 'efrac_mc'] if ('reso2' in sas.keys()) and mc_errors else []):
        sas[k] = np.array(sas[k])
    return sas

def to_correction_lib(sas: Union[Dict, str, Path], dir_results:Union[Path, str]='./tmp/', dset_name='DSET', 
                      cset_name='CSET', cset_description:str=None, cset_version=1, 
                      cat_latex=None) -> Tuple[str, List[str]]:
    """Create a correction lib files based 

    Args:
        sas (Dict, ): dictionnary with scales and smearings
        dir_results (Path, optional): directory. Defaults to Path('./tmp/').
        dset_name (str, optional): dataset nickname. Defaults to 'DSET'.
        cset_name (str, optional): correction saet name. Defaults to 'CSET'.
        cset_description (str, optional): description of the correction set. Defaults to None.
        cset_version (int, optional): correction set version. Defaults to 1.
        cat_latex (Dict, optional): dictionnary with a description of variables in categories. Defaults to None.
    """
    import correctionlib.schemav2 as cs
    import rich
    if not isinstance(sas, Dict):
        sas = json_safe_load(sas)

    dir_results = Path(dir_results)
    print(f'---------------------------------------')
    print(f'---- Create correction lib')
    print(f'---------------------------------------')
    if cat_latex is None:
        cat_latex = {key: key for key in sas['categories']}

    def create_correction(cset_name, out_name, measures):
        def create_bin_corr(s_name):
            # flow = cs.Binning(['error'] + ['clamp'] * len(sas['categories']))
            # flow = {'syst': 'error'} | {k: 'clamp' for k in sas['categories'].keys()}
            
            flow = 'clamp'
            return cs.MultiBinning( nodetype='multibinning',
                                inputs=list(sas['categories'].keys()),
                                edges=list(sas['categories'].values()),
                                content=measures[s_name],
                                flow=flow)

        return cs.Correction(name = cset_name, 
                                version = cset_version,
                                description=cset_description,
                                inputs = [cs.Variable(name="syst", type="string", description="Systematic type")] +
                                         [cs.Variable(name=dim, type='real', description=cat_latex[dim]) for dim in sas['categories']],
                                output = cs.Variable(name=out_name, type="real", description=f"EM {out_name} value"),
                                data = cs.Category(nodetype='category',
                                                   input='syst',
                                                   content=[cs.CategoryItem(key=k,value=create_bin_corr(k)) for k in measures.keys()]
                                                  )
                                )
        
    esr_rel = sas['eresp']/sas['resp']
    scales = {'scale': 1./sas['resp'], 'escale': esr_rel/sas['resp'], 'scale_up': 1./sas['resp']*(1 + esr_rel), 'scale_down': 1./sas['resp']*(1-esr_rel)}
    smears = {'smear': sas['reso'], 'esmear': sas['ereso'], 'smear_up': sas['reso']+sas['ereso'], 'smear_down': sas['reso']-sas['ereso'],
              'escale': esr_rel, 'scale_up': 1 + esr_rel, 'scale_down': 1 - esr_rel }
    if 'reso2' in sas.keys():
        smears |= {'reso2': sas['reso2'], 'ereso2': sas['ereso2'], 'reso2_up': sas['reso2']+sas['ereso2'], 'reso2_down': sas['reso2']-sas['ereso2'],
                  'mu': sas['mu'],  'emu': sas['emu'], "mu_up": sas['mu']+sas['emu'], "mu_down": sas['mu']-sas['emu'],
                  'frac': sas['frac'],'efrac': sas['efrac'], 'frac_up': sas['frac']+sas['efrac'], 'frac_down': sas['frac']-sas['efrac']}
                  
    c_scale = create_correction(f'EGMScale_{cset_name}_{dset_name}'       , 'scale', scales) # only scales to apply to data
    c_smear = create_correction(f'EGMSmearAndSyst_{cset_name}_{dset_name}', 'smear', smears) # MC parameters
    
    rich.print(c_scale)
    rich.print(c_smear)
    cset = cs.CorrectionSet(schema_version=2,
                            description=cset_description,
                            corrections=[c_scale, c_smear]
                            )
    dir_results.mkdir(exist_ok=True)
    file_out = dir_results / f"EGMScalesSmearing_{dset_name}.v{cset_version}.json"
    print(f"Writing out correction lib file: {file_out}.gz")
    import gzip
    with gzip.open(f"{file_out}.gz", "wt") as fout:
        fout.write(cset.model_dump_json(indent=1,exclude_unset=True))
    return f"{file_out}.gz", list(sas['categories'].keys()),

def compute_syst_from_jsons(nominal:Union[Dict, str, Path], syst_jsons:List[Union[Dict, str, Path]], fit_2g:Union[Dict, str, Path]=None,
                            scale_flat_syst:float=0.0, smear_flat_syst:float=0.0) -> Tuple[str, List[str]]:
    """Compute the systematic uncertainties from the json files

    Args:
        nominal (Union[Dict, str, Path]): nominal json file
        syst_jsons (List[Union[Dict, str, Path]]): list of json files with systematic variations
        fit_2g (Union[Dict, str, Path], optional): json file with the parameters of the double gaussian fit. Defaults to None.
        scale_flat_syst (float, optional): flat scale systematic uncertainty. Defaults to 0.0.
        smear_flat_syst (float, optional): flat smear systematic uncertainty. Defaults to 0.0.
    
    Returns:
    """ 
    if not isinstance(nominal, Dict):
        nominal = json_safe_load(nominal, mc_errors=True) 
    else:
        raise ValueError('Missing nominal json file')
    
    eresp = nominal['eresp']**2
    ereso = nominal['ereso']**2

    eresp += scale_flat_syst**2
    ereso += smear_flat_syst**2

    for syst in syst_jsons:
        if not isinstance(syst, Dict):
            syst = json_safe_load(syst)
        eresp += (syst['resp']-nominal['resp'])**2
        ereso += (syst['reso']-nominal['reso'])**2

    

    if fit_2g:
        if not isinstance(fit_2g, Dict):
            fit_2g = json_safe_load(fit_2g)
        import copy
        nominal_eresp = copy.deepcopy(eresp)
        nominal_ereso = copy.deepcopy(ereso)
        r = fit_2g['resp']
        s1 = fit_2g['reso']
        s2 = fit_2g['reso2']
        mu = fit_2g['mu']
        f = fit_2g['frac']

        fit_2g['mean2G'] = (f + (1-f)*mu)*r
        fit_2g['std2G'] = np.sqrt(f*s1**2 + (1-f)*(mu**2)*(s2**2) + f*(1-f)*((mu-1)**2)) * r

        eresp += (nominal['resp'] - fit_2g['mean2G'])**2
        ereso += (nominal['reso'] - fit_2g['std2G'])**2

        fit_2g['eresp'] = np.sqrt(nominal_eresp)
        fit_2g['ereso'] = np.sqrt(nominal_ereso)

        fit_2g['eresp_mc'] = nominal['eresp_mc']
        fit_2g['ereso_mc'] = nominal['ereso_mc']

    nominal['eresp'] = np.sqrt(eresp)
    nominal['ereso'] = np.sqrt(ereso)

    if fit_2g:
        return nominal, fit_2g
    else:
        return nominal 
   

def apply_corrlib_df(df: pd.DataFrame, cset_vars:List[str], cset_name:str, cset_file:str, syst_name:str, mll_name='mass', ptl_name='pt'):
    """Apply the correction lib to a pandas dataframe

    Args:
        df (pd.DataFrame): dataframe with the variables to correct
        cset_vars (List[str]): list of the variables used in the correction
        cset_name (str): name of the correction set
        cset_file (str): name of the correction file
        syst_name (str): type of correction to apply: scale or smear
        mll_name (str, optional): dilepton mass variable name. Defaults to 'mass'.
        ptl_name (str, optional): lepton tranverse momentum variable name. Defaults to 'pt'.
    """
    import correctionlib
    cset = correctionlib.CorrectionSet.from_file(cset_file)
    try:
        corr = cset[cset_name]
    except IndexError:
        corr = cset.compound[cset_name]

    lep_names = [f'{name}1' for name in cset_vars]
    lep_names = [name.split('1')[0] for name in df.columns.intersection(lep_names)]
    for i_ele in [1, 2]:
        lep_vars = [df[f'{var}{i_ele}' if var in lep_names else var] for var in cset_vars]

        scales = corr.evaluate(*([syst_name] + lep_vars))
        if syst_name == 'smear':
            try:
                scales *= corr.evaluate(*(['stdnormal'] + lep_vars))
                print('Smearing using the included random number generator')
            except:
                scales = np.random.normal(1, scales)
                print('Smearing using the numpy random')

        df[mll_name] *= np.sqrt(scales)
        df[f'{ptl_name}{i_ele}'] *= scales


# ------------------------------------------------------------------------------------------
# -- Potentially fit the results of the SAS when they are computed vs pT
# -- This is done to avoid spikes in the pT spectrum (+ smoothen the result)
# ------------------------------------------------------------------------------------------
class FuncStr:
    def __init__(self, expression: str):
        """This class generates a numpy verctorized function based on a string expression
            Parameters for fits are possible with [], e.g. par0 == [0]
        Args:
            expression (str): analytical expression of the form "np.sqrt([1]*x + [0])

        Raises:
            SyntaxError: checkout that the syntax is correct (only [] for now)
        """
        if expression.count("[") != expression.count("]"):
            raise SyntaxError("Missing a stating or closing braket")
        self.n_par = 1 + expression.count("[")
        self.expression = expression

    def replace_par(self, *par):
        expression = self.expression
        for p_name, p_val in zip([f"[{ip}]" for ip in range(self.n_par-1)], par):
            expression = expression.replace(p_name, f'{p_val:.10f}')
        return expression

    def tformula(self, *par) -> str:
        """Return a ROOT type TFormula replacing parameters with inputs pars

        Returns:
            str: TFormula expression
        """
        expression = self.expression
        for p_name, p_val in zip([f"[{ip}]" for ip in range(self.n_par-1)], par):
            expression = expression.replace(p_name, f'{p_val:.10f}')
        expression = expression.replace('np.', '')
        expression = expression.replace('power', 'pow')
        return expression
    
    def __call__(self, x, *par):
        return np.frompyfunc(lambda x, *par: eval(self.replace_par(*par)), self.n_par, 1)(x, *par)


resp_exp = "[0] * np.power(min(x, 160)/45., [1]) + [2]"  # fix corr for pT > 160 GeV
reso_exp = "np.sqrt(np.abs(np.power([0], 2) + [1]/min(x, 160)))"   # fix corr for pT > 160 GeV

pt_fit_resp = FuncStr(resp_exp)
pt_fit_reso = FuncStr(reso_exp)

p0_resp = [-0.02, -1., 1.0]
pr_resp = [[-0.9, 0.5], [-5, 5], [0.5, 1.5]]
p0_reso = [0.01, 0.0]
pr_reso = [[1e-3, 0.1], [-0.05, 0.05]]


def do_pt_fit(axval, avar, eavar, func:FuncStr, p0, p_range, 
           do_plot=False, min_pt=25, max_pt = 160, width_pt=0.5):
    
    minimizer="BFGS"

    fit_result = [None] * avar.shape[-1]
    if do_plot:
        n_row = int(np.ceil(avar.shape[-1]/3.))
        fig, ax = plt.subplots(3, n_row, figsize=(9, 3*n_row) )
        ax = ax.flatten()
    
    fine_pt_bining = np.linspace(min_pt, max_pt, int((max_pt-min_pt)/width_pt) + 1 )
    xxval = (fine_pt_bining[1:] + fine_pt_bining[:-1])*0.5
    y_fine = []
    for i_dim in np.arange(avar.shape[-1]):
        fitter = FFitter(p0=p0, p_range=p_range, minimizer=minimizer)
        par, epar = fitter.fit(axval, avar[:, i_dim], eavar[:, i_dim], func)
        if par is None:
            fitter = FFitter(p0=p0, p_range=p_range, minimizer='Nelder-Mead')
            par1, epar1 = fitter.fit(axval, avar[:, i_dim], eavar[:, i_dim], func) 
            print(" * refit1", par1)
            fitter = FFitter(p0=par1, p_range=p_range, minimizer=minimizer)
            par, epar = fitter.fit(axval, avar[:, i_dim], eavar[:, i_dim], func) 
            print(" * refit2", par)
            if par is None:
                par, epar = (par1, epar1)
        y_fine.append(func(xxval, *par))
        if do_plot:
            plt.sca(ax[i_dim])
            cms.draw(axval, avar[:, i_dim], yerr=eavar[:, i_dim], option ='E')
            ax[i_dim].plot(xxval, y_fine[-1], color=ax[i_dim].lines[-1].get_color(), ls='--')
        fit_result[i_dim] = func.tformula(*par)

    return fit_result 


def pt_smoothing(json_file:Union[str, Path, Dict], dim_to_fit='pt', do_plot=False) -> Tuple[Dict, Dict]:
    """Function to fit the pT-dependent sas to make them smooth

    Args:
        json_file (Union[str, Path]): name of the json file containing the result of the sas fit
        dim_to_fit (str, optional): name of the pt dimension. Defaults to 'pt'.

    Returns:
        Dict: new sas dictionnary
    """
    params = parameters_from_json(json_file)
    if params.get('pt', None) is not None:
        xval = params['pt']
    else:
        return {}
        
    fit_vres = {}
    fit_eres = {}
    for sas_type in ['resp', 'reso']:
        var, evar, evar_mc = params[sas_type], params[f'e{sas_type}'], params[f'e{sas_type}_mc']
        
        # -- find the dimension with pt
        idim_pt = np.where(params['dims'] == dim_to_fit)[0][-1]
        d_perm = [idim_pt] + [idim for idim, d in enumerate(params['dims']) if d != dim_to_fit]
        var = tf.transpose(var, perm=d_perm).reshape(var.shape[idim_pt], -1)
        evar = tf.transpose(evar, perm=d_perm).reshape(evar.shape[idim_pt], -1)
        evar_mc = tf.transpose(evar_mc, perm=d_perm).reshape(evar.shape[idim_pt], -1)

        p0 = p0_resp if sas_type == 'resp' else p0_reso
        pr = pr_resp if sas_type == 'resp' else pr_reso
        pt_fit = pt_fit_resp if sas_type == 'resp' else pt_fit_reso

        var = var.numpy()
        evar = evar.numpy()
        var_up = var + evar
        if sas_type == 'resp':
            var = 1 / var
            var_up = 1 / (var - evar)
        res = do_pt_fit(xval, var, evar, pt_fit, p0, pr, do_plot=do_plot)
        res_up = do_pt_fit(xval, var_up, evar, pt_fit, p0, pr)
        fit_vres[sas_type] = res
        fit_eres[sas_type] = res_up

    eresp_abs = f'abs({fit_vres["resp"]} - {fit_vres["resp"]})'
    ereso_abs = f'abs({fit_vres["reso"]} - {fit_vres["reso"]})'
    eresp_rel = f'abs({fit_vres["resp"]}/{fit_vres["resp"]} - 1)'

    fit_res_dt = {'scale': fit_vres['resp'], 'escale': eresp_abs}
    fit_res_mc = {'smear': fit_vres['reso'], 'escale': eresp_rel, 'esmear': ereso_abs}
    return fit_res_dt,  fit_res_mc


def create_pt_corrector(sas: Union[Dict, str, Path], dir_results:Union[Path, str]='./tmp/', dset_name='DSET', cset_name='CSET',
                        cset_description:str=None, cset_version=1, cat_latex=None, pt_name='pt'):
    """Create a correction lib files based 

    Args:
        sas (Dict, ): dictionnary with scales and smearings
        dir_results (Path, optional): directory. Defaults to Path('./tmp/').
        dset_name (str, optional): dataset nickname. Defaults to 'DSET'.
        cset_name (str, optional): correction saet name. Defaults to 'CSET'.
        cset_description (str, optional): description of the correction set. Defaults to None.
        cset_version (int, optional): correction set version. Defaults to 1.
        cat_latex (Dict, optional): dictionnary with a description of variables in categories. Defaults to None.
    """
    import correctionlib.schemav2 as cs
    import rich
    print(f'--------------------------------------------')
    print(f'---- Create correction lib with PT SMOOTHING')
    print(f'--------------------------------------------')
    # -- do fits
    fit_dt, fit_mc = pt_smoothing(sas)

    if not isinstance(sas, Dict):
        sas = json_safe_load(sas)
    
    if cat_latex is None:
        cat_latex = {key: key for key in sas['categories']}
    dir_results = Path(dir_results)

    categories = sas['categories'].copy()
    categories.pop(pt_name)
    edges = list(categories.values())
    n_val = np.prod([len(e)-1 for e in edges])

    def get_correction(sas_type, fit_results):
        input_cats = [cs.Variable(name=dim, type='real', description=cat_latex[dim]) for dim in  categories.keys()]
        
        def get_pt_corr(expression, flow='clamp'):
            content=[cs.Formula(nodetype="formula", variables=[pt_name], parser="TFormula", expression=expression[ibin]) for ibin in range(n_val)]
            return cs.MultiBinning(nodetype='multibinning',
                                   inputs=list(categories.keys()),
                                   edges=edges,
                                   content=content,
                                   flow=flow)  
            
        cset = cs.Correction(name=f"EGMScale_{cset_name}_{dset_name}" if sas_type == 'scale' else f"EGMSmearAndSyst_{cset_name}_{dset_name}" ,
                            version = cset_version,
                            description=cset_description,
                            inputs = [cs.Variable(name="syst", type="string", description="Systematic type"),
                                      cs.Variable(name=pt_name, type="real", description="p_T in GeV")] + input_cats ,
                            output = cs.Variable(name=sas_type, type="real", description=f"EM {sas_type} value"),
                            data = cs.Category(nodetype='category',
                                               input='syst',
                                               content=[cs.CategoryItem(key=k,value=get_pt_corr(v)) for k, v in fit_results.items()],
                                               default=get_pt_corr(fit_results[sas_type]))
                            )
        return cset

    c_scale = get_correction('scale', fit_dt)
    c_smear = get_correction('smear', fit_mc)
    rich.print(c_scale)
    rich.print(c_smear)
    
    cset = cs.CorrectionSet(schema_version=2,
                            description=cset_description,
                            corrections=[c_scale, c_smear]
                            )
    dir_results.mkdir(exist_ok=True)
    file_out = dir_results / f"EGMScalesSmearing_{dset_name}_SMOOTH.v{cset_version}.json"
    print(f"Writing out correction lib file: {file_out}.gz")
    import gzip
    with gzip.open(f"{file_out}.gz", "wt") as fout:
        fout.write(cset.model_dump_json(indent=1,exclude_unset=True))
    return c_scale


def ijazz_sas_smoothing():
    """entry point for the ijazz_sas_smoothing command
    """
    import argparse
    import sys
    parser = argparse.ArgumentParser(description=f'IJazZ Scale and Smearing fit')
    parser.add_argument('json_file', type=str, help='name of json file')
    parser.add_argument('--dset', type=str, default='DSET', help="name of the dataset")
    parser.add_argument('--cset', type=str, default='CSET', help="name of the correction set")
    parser.add_argument('--description', type=str, default=None, help="cset description")
    parser.add_argument('-v', '--vers', type=int, default=1, help="cset version")
    parser.add_argument('-o', '--outdir', type=str, default="./", help="output directory")
    args = parser.parse_args(sys.argv[1:])

    create_pt_corrector(args.json_file, dir_results=args.outdir, dset_name=args.dset, cset_name=args.cset,
                        cset_description=args.description, cset_version=args.vers) 







__doc__ = "This module allows to create a simple MC simulation of the Z->ee decay" \
          "with a gaussian smearing and scale shift that can be tune w/r " \
          "to a property x which is generated in the range [0., 100.]" \
          "Author: fabrice.couderc@cea.fr" \
          "Revision: v1 - 21/03/2024"


from ijazz.dtypes import floatzz
import tensorflow as tf
import tensorflow_probability as tfp
import numpy as np, pandas as pd


mass_z = tf.constant(91.1876, dtype=floatzz)
gamma_z = tf.constant(2.4952, dtype=floatzz)
win_z_mc = (50, 130)
win_z_dt = (70, 110)


def decalibrate(x):
    """
    Decalibrate the response of the events for the property x.

    Args:
        x (np.ndarray): Input property.

    Returns:
        tf.Tensor: Decalibrated response.
    """
    return 1 + x/1000 * tf.sin(x/5)


def oversmearing(x):
    """Applies additional oversmearing based on the property x.

    Args:
        x (np.ndarray): Input property.

    Returns:
        np.ndarray: Oversmearing values.
    """
    return np.ones(x.shape) * 0.020


# -- oversmearing function more realistic
def oversmearing2(x):
    # return 0.015 * np.ones(tf.shape(x))
    return 0.015 * (1 + (x / 100) ** 2)


def dataset(do_decal, n_toys=100, decalibrate_f=decalibrate, oversmearing_f=oversmearing,
            smear_property=False, prop_gaus=False, return_smear=False):
    """
    Create the dataset.

    Args:
        prop_gaus (bool): If True, generate properties with a Gaussian distribution (mean=45, sigma=25).
        do_decal (bool): If True, decalibrate the Z BW events using the `decalibrate_f` and `oversmearing_f` functions.
        n_toys (int): Number of Monte Carlo toys to generate.
        decalibrate_f (Callable): Function to apply scale decalibration.
        oversmearing_f (Callable): Function to apply oversmearing.
        smear_property (bool): If True, smear the properties `x1` and `x2` according to the smearing in the mass.
        prop_gaus (bool): If True, generate properties with a Gaussian shape.
        return_smear (bool): If True, return smeared values.

    Returns:
        Tuple[tf.Tensor, tf.Tensor, tf.Tensor]: A tuple containing the smeared mass (`mee_smear`), 
        and the properties `x1` and `x2` of the two electrons. If `return_smear` is True, also return 
        the smearing factors `s1` and `s2`.
    """

    mee_bw = tf.cast(tfp.distributions.Cauchy(mass_z, gamma_z / 2.).sample(n_toys), floatzz)
    # e1_smear = tf.ones(n_toys, dtype=floatzz)
    # e2_smear = tf.ones(n_toys, dtype=floatzz)
    e1_smear = np.random.normal(1, 0.02, n_toys)
    e2_smear = np.random.normal(1, 0.02, n_toys)
    mee_smear = mee_bw * tf.sqrt(e1_smear * e2_smear)
    prop_min = 0
    prop_max = 100

    # -- generate properties with a larger range than [prop_min, prop_max], so it can be randomized
    if prop_gaus:
        x1 = tf.cast(tfp.distributions.Normal(45, 18).sample(n_toys), floatzz)
        x2 = tf.cast(tfp.distributions.Normal(45, 18).sample(n_toys), floatzz)
    else:
        x1 = tf.cast(tfp.distributions.Uniform(prop_min-10, prop_max+10).sample(n_toys), floatzz)
        x2 = tf.cast(tfp.distributions.Uniform(prop_min-10, prop_max+10).sample(n_toys), floatzz)

    if do_decal:
        corr1 = decalibrate_f(x1)
        corr2 = decalibrate_f(x2)
        e1_os = np.random.normal(1, oversmearing_f(x1))
        e2_os = np.random.normal(1, oversmearing_f(x2))
        mee_smear *= tf.sqrt(corr1 * corr2 * e1_os * e2_os)
        if smear_property:
            x1 *= corr1 * e1_os
            x2 *= corr2 * e2_os
        s1 = corr1 * e1_os
        s2 = corr2 * e2_os

    win_z = win_z_mc
    if do_decal:
        win_z = win_z_dt
        sel_dt = (mee_smear >= win_z[0]) & (mee_smear <= win_z[1])
        mee_smear = mee_smear[sel_dt]
        x1 = x1[sel_dt]
        x2 = x2[sel_dt]
        s1 = s1[sel_dt]
        s2 = s2[sel_dt]

    # -- force x1 and x2 to be in [0, 100] after potential x1 and x2 smearing
    sel = (x1 >= prop_min) & (x2 >= prop_min) & (x1 < prop_max) & (x2 < prop_max)
    x1 = x1[sel]
    x2 = x2[sel]
    if do_decal:
        s1 = s1[sel]
        s2 = s2[sel]
    mee_smear = mee_smear[sel]
    if return_smear :
        return tf.clip_by_value(mee_smear, *win_z), x1, x2, s1, s2
    
    else :
        return tf.clip_by_value(mee_smear, *win_z), x1, x2


# --------------------------------------------------------------------------
# - Next section adresses the Pythia simulation (decalibration and smearing)
# --------------------------------------------------------------------------

def tree_to_df(tree, random_R9=True, use_fabrice=False):
    mc_m_z, mc_pt_z, mc_pt_e1_true, mc_pt_e2_true, mc_y_e1_true, mc_y_e2_true, mc_phi_e1_true, mc_phi_e2_true = \
        tree_to_df_fabrice(tree) if use_fabrice else tree_to_df_paul(tree)

    if random_R9:
        r91 = np.random.normal(0.94, 0.1, len(mc_pt_e1_true))
        r92 = np.random.normal(0.94, 0.1, len(mc_pt_e1_true))
    
    sel = (mc_m_z > 50) & (mc_pt_z > 0) & (mc_pt_e1_true > 10) & (mc_pt_e2_true > 10) # -- minimum pT cut added 
    df = pd.DataFrame({'mee_true': mc_m_z[sel], 'pT_Z': mc_pt_z[sel], 
                        'pT1_true': mc_pt_e1_true[sel], 'pT2_true': mc_pt_e2_true[sel],
                        'y1': mc_y_e1_true[sel], 'abs_y1': abs(mc_y_e1_true[sel]),
                        'y2': mc_y_e2_true[sel], 'abs_y2': abs(mc_y_e2_true[sel]),
                        'phi1': mc_phi_e1_true[sel], 'phi2': mc_phi_e2_true[sel],
                        'R91': r91[sel], 'R92': r92[sel],})

    return df

def tree_to_df_paul(tree):
    mc_pt_e_i_true = tree["pT_e_i"].array(library="np")
    mc_y_e_i_true = tree["y_e_i"].array(library="np")
    mc_phi_e_i_true = tree["phi_e_i"].array(library="np")
    mc_id_e_i_true = tree["id_e_i"].array(library="np")
    mc_m_Z = tree["m_Z"].array(library="np")
    mc_m_Z_true = mc_m_Z[mc_m_Z>0]
    mc_pT_Z = tree["pT_Z"].array(library="np")
    mc_pT_Z = mc_pT_Z[mc_m_Z>0]

    #electrons
    mc_pt_e1_true = mc_pt_e_i_true[mc_id_e_i_true==11]
    mc_y_e1_true = mc_y_e_i_true[mc_id_e_i_true==11]
    mc_phi_e1_true = mc_phi_e_i_true[mc_id_e_i_true==11]

    #positrons
    mc_pt_e2_true = mc_pt_e_i_true[mc_id_e_i_true==-11]
    mc_y_e2_true = mc_y_e_i_true[mc_id_e_i_true==-11]
    mc_phi_e2_true = mc_phi_e_i_true[mc_id_e_i_true==-11]

    return mc_m_Z, mc_m_Z_true, mc_pt_e1_true, mc_pt_e2_true, mc_y_e1_true, mc_y_e2_true, mc_phi_e1_true, mc_phi_e2_true

def tree_to_df_fabrice(tree):
    mc_pt_e_i_true = tree["pt_top"].array(library="np")
    mc_y_e_i_true = tree["eta_top"].array(library="np")
    mc_phi_e_i_true = tree["phi_top"].array(library="np")

    #electrons
    i_ele = 0
    mc_pt_e1_true = mc_pt_e_i_true[:, i_ele]
    mc_y_e1_true = mc_y_e_i_true[:, i_ele]
    mc_phi_e1_true = mc_phi_e_i_true[:, i_ele]

    #positrons
    i_ele = 1
    mc_pt_e2_true = mc_pt_e_i_true[:, i_ele]
    mc_y_e2_true = mc_y_e_i_true[:, i_ele]
    mc_phi_e2_true = mc_phi_e_i_true[:, i_ele]

    mc_m_z = np.sqrt(2*mc_pt_e1_true*mc_pt_e2_true*
            (np.cosh(mc_y_e1_true-mc_y_e2_true) - 
            np.cos(mc_phi_e1_true-mc_phi_e2_true)))

    mc_pt_z = np.sqrt(mc_pt_e1_true**2 + mc_pt_e2_true**2 + 
            2*mc_pt_e1_true*mc_pt_e2_true* 
            np.cos(mc_phi_e1_true-mc_phi_e2_true))

    return mc_m_z, mc_pt_z, mc_pt_e1_true, mc_pt_e2_true, mc_y_e1_true, mc_y_e2_true, mc_phi_e1_true, mc_phi_e2_true

def detector_smearing(x1, x2, mee, sigma=0.02):
    """Adds Gaussian smearing with resolution `sigma` to `x1`, `x2`, and `mee`.

    Args:
        x1 (np.ndarray): Variable to apply the smearing (first electron).
        x2 (np.ndarray): Variable to apply the smearing (second electron).
        mee (np.ndarray): Variable to apply the smearing (dilepton system).
        sigma (float, optional): Gaussian resolution. Defaults to 0.02.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Smeared variables `x1_smear`, `x2_smear`, and `mee_smear`.
    """
    e1_os = np.random.normal(1, sigma, len(x1))
    e2_os = np.random.normal(1, sigma, len(x1))
    mee_smear = mee * np.sqrt(e1_os * e2_os)
        
    x1_smear = x1 * e1_os
    x2_smear = x2 * e2_os
    

    return x1_smear, x2_smear, mee_smear

def decal_and_os(x1, x2, pT1, pT2, mee, decalibrate_f=decalibrate, oversmearing_f=oversmearing, decal_kwargs={}, os_kwargs={}):
    """Decalibrate and add oversmearing to pT1, pT2, and mee.

    Args:
        x1 (tuple): Variables to use as input for the decalibrate and oversmearing functions (first lepton).
        x2 (tuple): Variables to use as input for the decalibrate and oversmearing functions (second lepton).
        pT1 (float): Variable to apply the decalibration and oversmearing (first lepton).
        pT2 (float): Variable to apply the decalibration and oversmearing (second lepton).
        mee (float): Variable to apply the decalibration and oversmearing (dilepton system).
        decalibrate_f (Callable): Function to decalibrate the pT. Defaults to `decalibrate`.
        oversmearing_f (Callable): Function to oversmear the pT. Defaults to `oversmearing_all`.

    Returns:
        Tuple[float, float, float, float]: pT1_smear, pT2_smear, mee_smear, mee_os (mee with only oversmearing).
    """
    corr1 = decalibrate_f(*x1, **decal_kwargs)
    corr2 = decalibrate_f(*x2, **decal_kwargs)
    e1_os = np.random.normal(1, oversmearing_f(*x1, **os_kwargs))
    e2_os = np.random.normal(1, oversmearing_f(*x2, **os_kwargs))
    mee_os = mee * np.sqrt(e1_os * e2_os)
    mee_smear = mee * np.sqrt(corr1 * corr2 * e1_os * e2_os)
    
    pT1_smear = pT1* corr1 * e1_os
    pT2_smear = pT2* corr2 * e2_os

    return pT1_smear, pT2_smear, mee_smear, mee_os


import numpy as np
import pandas as pd

def categorize(df: pd.DataFrame, category_dict: dict, cut="", prefix='cat', var_prefixes=None, var_suffixes=['1','2']) -> np.ndarray:
    """Create the categories for both lepton based on a dataframe containing the categorisation variable per lepton
    under the form var1, var2 for lepton 1 and lepton 2, the name of the variables to categorize upon are specified in the 
    the category dictionnary along with the corresponding bining.

    Args:
        df (pd.DataFrame): input dataframe containing the variable to categorize
        category_dict (dict): dictionary for categorisation, e.g. {'pt': [25, 50, 100], 'abs_eta': [0, 1, 2]}
        cut (str, optional): cut to be use if df.eval(cut) to apply a selection. Defaults to "".
        prefix (str, optional): prefix used for the categorisation. Defaults to 'cat'.

    Returns:
        np.ndarray: array with the category numbers that have been created
    """
 
    n_cat_bins = [len(bin)-1 for bin in category_dict.values()]
    categories = np.arange(np.prod(n_cat_bins)).reshape(*n_cat_bins)
    selection = True
    idx = slice(0, None) if cut == "" else df.eval(cut)
    if var_prefixes is None:
        var_prefixes = ['','']
    else:
        var_suffixes = ['','']
    for name, bins in category_dict.items():
        for ele_suffix, ele_prefix in zip(var_suffixes, var_prefixes):
            df[f'{prefix}_{ele_prefix}{name}{ele_suffix}'] = np.int32(-1)
            df.loc[idx, f'{prefix}_{ele_prefix}{name}{ele_suffix}'] = (np.digitize(df.loc[idx, f"{ele_prefix}{name}{ele_suffix}"], bins) - 1).astype(np.int32)
            selection &= (df[f'{prefix}_{ele_prefix}{name}{ele_suffix}'] >= 0) & (df[f'{prefix}_{ele_prefix}{name}{ele_suffix}'] < len(bins) - 1)

    for i_ele, (ele_suffix, ele_prefix) in enumerate(zip(var_suffixes, var_prefixes)):
        i_ele += 1
        df[f'{prefix}{i_ele}'] = -1
        df.loc[selection, f'{prefix}{i_ele}'] = categories[tuple([df.loc[selection, f"{prefix}_{ele_prefix}{name}{ele_suffix}"] for name in category_dict.keys()])]
        df[f'{prefix}{i_ele}'] = df[f'{prefix}{i_ele}'].astype(np.int32)
    return categories


from ijazz.ScaleAndSmearing import IJazZSAS
from ijazz.sas_utils import parameters_from_json
import matplotlib.pyplot as plt
import tensorflow as tf
from matplotlib.pyplot import Figure, Axes
import cms_fstyle as cms
import numpy as np
import pandas as pd
from typing import Union, List, Dict, Tuple
from pathlib import Path
import copy


def ijazz_plot_results():
    """entry point for the ijazz_plot command
    """
    import argparse, sys, json
    parser = argparse.ArgumentParser(description=f'IJazZ json result plotter')
    parser.add_argument('json_file', type=str, help='json file with the fit results')
    parser.add_argument('--resp', type=float, nargs='+', default=(), help='response y range')
    parser.add_argument('--reso', type=float, nargs='+', default=(), help='resolution y range')
    parser.add_argument('-x', '--x_range', type=float, nargs='+', default=(), help='x range')
    parser.add_argument('--latex', type=str, default=None, help='json file with latex categories')
    parser.add_argument('-o', '--out', type=str, default=None, help='name of the output file')
    parser.add_argument('-d', '--dims', type=str, nargs='+', default=None, help='dimensions to plot')
    parser.add_argument('--ncol', type=int, default=1, help='number of columns for legend')
    parser.add_argument('--leg_fontsize', type=str, default='medium', help='fontsize for legend')
    parser.add_argument('--leg_title_fontsize', type=str, default='large', help='title fontsize for legend')
    parser.add_argument('-H','--hide', action='store_true', help='hide legend')

    args = parser.parse_args(sys.argv[1:])
    print(args)

    cat_latex = None
    if args.latex:
        with open(args.latex, 'r') as flatex:
            cat_latex=json.load(flatex)
        
    figs, ax = plot_results_from_json(args.json_file, dims=args.dims, x_range=args.x_range, resp_range=args.resp, reso_range=args.reso, cat_latex=cat_latex,
                                      hide_legend=args.hide, leg_ncol=args.ncol, leg_fontsize=args.leg_fontsize, leg_title_fontsize=args.leg_title_fontsize)
    plt.show()
    if args.out:
        out = Path(args.out)
        out.parent.mkdir(parents=True, exist_ok=True)
        print(f' --> Save plots to {out}')
        for ifig, fig in enumerate(figs):
            fig.savefig(out.with_suffix(f'.fig{ifig}' + out.suffix))


def plot_results_from_json(json_file: Union[str, Path, Dict], dims: List=None, x_range=(), x_unit='', 
              cat_latex:Dict=None, resp_range=(), reso_range=(), hide_legend=False, 
              jsons_mode='', param_to_plot=None, **kwargs) -> Tuple[List[Figure], Axes]:
    """Create a list of Axes to plot the results in the json file.
    the dimensions to plot can be specified, additional dimension are averaged over.
    Up to 3 dimensions can be plotted.

    Args:
        json_file (Union[str, Path, Dict]): json file containing the results of the IJazZ SAS fit or Dict from the json file
        dims (List, optional): name dimensions to plot (as in the categories). Defaults to None.
        x_range (tuple, optional): range of the main variable (dimension 0). Defaults to ().
        x_unit (str, optional): usint of the main varaible. Defaults to ''.
        cat_latex (Dict, optional): dictionnary mapping var name (from categories) into a latex name. Defaults to None.
        resp_range (tuple, optional): y-range of the resp axis. Defaults to ().
        reso_range (tuple, optional): y-range if the reso axis. Defaults to ().
        leg_ncol (int, optional): number of columns in the legend. Defaults to 1.
        hide_legend (bool, optional): hide the legend. Defaults to False.
        jsons_mode (str, optional): select plotting mode between 'compare' and 'ratio'. Defaults to ''.
        param_to_plot (str, optional): parameter to plot (resp or reso) if compare_jsons. Defaults to None.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the plot.
    """
    # -- get the parameters
    if compare_jsons:= jsons_mode == 'compare':
        params_dict = {}
        for key, values in json_file.items():
            params_dict[key] = values if isinstance(values, Dict) else parameters_from_json(values)
        # params = params_dict[key]
        json_names = list(params_dict.keys())
        params_list = list(params_dict.values())
        params = params_list[0]

    elif jsons_mode == 'ratio':
        json_nominal = json_file.get('nominal')
        json_syst = json_file.get('syst')
        if json_nominal is None or json_syst is None:
            raise KeyError('the json file must contain a nominal and syst key')

        params_nominal = json_nominal if isinstance(json_nominal, Dict) else parameters_from_json(json_nominal)
        params_syst = json_syst if isinstance(json_syst, Dict) else parameters_from_json(json_syst)
        params = copy.deepcopy(params_syst)
        # params = params_nominal
        # params['resp'] = params_syst['resp'][2:,:,:] / params_nominal['resp']
        # params['eresp'] = params_syst['eresp'][2:,:,:] / params_nominal['resp']
        # params['eresp_mc'] = params_syst['eresp_mc'][2:,:,:] / params_nominal['resp']
        # params['reso'] = params_syst['reso'][2:,:,:] / params_nominal['reso']
        # params['ereso'] = params_syst['ereso'][2:,:,:] / params_nominal['reso']
        # params['ereso_mc'] = params_syst['ereso_mc'][2:,:,:] / params_nominal['reso']
        
        params['resp'] = params_syst['resp'] / params_nominal['resp']
        params['eresp'] = params_syst['eresp'] / params_nominal['resp']
        params['eresp_mc'] = params_syst['eresp_mc'] / params_nominal['resp']
        params['reso'] = params_syst['reso'] / params_nominal['reso']
        params['ereso'] = params_syst['ereso'] / params_nominal['reso']
        params['ereso_mc'] = params_syst['ereso_mc'] / params_nominal['reso']
        

    else:
        params = json_file if isinstance(json_file, Dict) else parameters_from_json(json_file)

    if not cat_latex:
        # -- define cat_latex if not given
        cat_latex = {key: key for key in params['dims']}

    if dims is None:
        # -- keep all dims by default
        dims = [key for key in params['dims']]
    
    for dim in dims:
        if dim not in cat_latex:
            cat_latex[dim] = dim

    if len(dims) > 3:
        # -- keep only the 3 first dimensions
        dims = dims[:3]

    dims_to_keep = [np.where(params['dims'] == d)[0][0] for d in dims]
    dims_to_average = [id for id, d in enumerate(params['dims']) if not d in dims]
    perm = dims_to_keep + dims_to_average
    dims_to_average = np.arange(len(dims_to_keep), len(perm))

    def reshape_param(param, par_name):
        par = tf.transpose(param[par_name], perm=perm)
        epar = tf.transpose(param[f'e{par_name}'], perm=perm)
        epar_mc = tf.transpose(param[f'e{par_name}_mc'], perm=perm)
        # avoid 0 division
        epar = tf.where(epar == 0, 1, epar)
        epar_mc = tf.where(epar_mc == 0, 1, epar_mc)

        par = tf.reduce_sum((par/ epar**2), axis=dims_to_average) / tf.reduce_sum((1 / epar**2), axis=dims_to_average)
        epar = tf.sqrt(1./ tf.reduce_sum((1 / epar**2), axis=dims_to_average))
        epar_mc = tf.sqrt(1./ tf.reduce_sum((1 / epar_mc**2), axis=dims_to_average))
        # put back errors to 0
        epar = tf.where(epar == 1, 0, epar)
        epar_mc = tf.where(epar_mc == 1, 0, epar_mc)
        shape = list(tf.shape(par).numpy()) + [1] * (3-tf.rank(par).numpy())
        return tf.reshape(par, shape), tf.reshape(epar, shape), tf.reshape(epar_mc, shape)

    if compare_jsons:

        resps, eresps, eresp_mcs = zip(*(reshape_param(param, 'resp') for param in params_list))
        resos, eresos, ereso_mcs = zip(*(reshape_param(param, 'reso') for param in params_list))
        resp, reso = np.stack(resps, axis=1), np.stack(resos, axis=1)
        eresp, ereso = np.stack(eresps, axis=1), np.stack(eresos, axis=1)
        eresp_mc, ereso_mc = np.stack(eresp_mcs, axis=1), np.stack(ereso_mcs, axis=1)
        
    
        if len(dims) < 3:
            resp = np.squeeze(resp, axis=-1)
            reso = np.squeeze(reso, axis=-1)
            eresp = np.squeeze(eresp, axis=-1)
            ereso = np.squeeze(ereso, axis=-1)
            eresp_mc = np.squeeze(eresp_mc, axis=-1)
            ereso_mc = np.squeeze(ereso_mc, axis=-1)
        
    else:
        resp, eresp, eresp_mc = reshape_param(params,'resp')
        reso, ereso, ereso_mc = reshape_param(params,'reso')

    # -- dim 0
    xbin = params['bins'][dims_to_keep[0]]
    xbin = tf.where(xbin == +np.inf, x_range[1] if len(x_range) > 1 else xbin[-2]*1.5, xbin)
    xbin = tf.where(xbin == -np.inf, x_range[0] if len(x_range) > 0 else xbin[1]/1.5, xbin)
    xval = 0.5*(xbin[1:] + xbin[:-1])
    xerr = 0.5*(xbin[1:] - xbin[:-1])
    var_ref = cat_latex[dims[0]] + x_unit

    # -- higher dimensions
    var1, var2 = None, None
    bin1, bin2 = [], []
    n_dim2 = 1
    if compare_jsons:
        json_names 
        if len(dims) > 1:
            var1 = cat_latex[dims[1]] 
            bin1 = params['bins'][dims_to_keep[1]]
            n_dim2 = len(bin1) - 1          

        if len(dims) > 2:
            var2 = cat_latex[dims[2]] 
            bin2 = params['bins'][dims_to_keep[2]]
            n_dim2 = len(bin2) - 1

            if len(params['bins'][dims_to_keep[1]]) - 1 >2:
                print('WARNING: too many bins for the 2th dimension, only the first 2 will be considered')
            
    else:
        if len(dims) > 1:
            var1 = cat_latex[dims[1]] 
            bin1 = params['bins'][dims_to_keep[1]]

        if len(dims) > 2:
            var2 = cat_latex[dims[2]] 
            bin2 = params['bins'][dims_to_keep[2]]
            n_dim2 = len(bin2) - 1

    # -- only maximum of 2 line per figure
    figs = []
    axes = []
    for ifig in range(n_dim2 // 2):
        fig, ax = plt.subplots(2, 2, figsize=(12, 4.5*2), squeeze=False)
        figs.append(fig)
        axes.append(ax)
    for ifig in range(n_dim2 % 2):
        fig, ax = plt.subplots(1, 2, figsize=(12, 4.5), squeeze=False)
        figs.append(fig)
        axes.append(ax)
    ax = np.concatenate(axes, axis=0)

    for i2 in range(resp.shape[-1]):
        leg_titles = []
        for i1 in range(len(params_list) if compare_jsons else resp.shape[-2]):
            label = f"${bin1[i1]:>5.4g} \leq${var1}$< {bin1[i1+1]:<5.4g}$" if var1 and not (compare_jsons) else json_names[i1] if compare_jsons else "fit" 
            leg_title = '' if not var2 else f"${bin2[i2]:>5.4g} \leq${var2}$< {bin2[i2+1]:<5.4g}$" 
            if compare_jsons and len(dims) != 1:
                leg_title += f"\n${bin1[min(i1,resp.shape[-2]-1) if len(dims)>2 else i2]:>5.4g} \leq${var1}$< {bin1[min(i1,resp.shape[-2]-1)+1 if len(dims)>2 else i2+1]:<5.4g}$"
                leg_titles.append(leg_title)

            def get_resp_reso(resp, reso, corr='resp'):
                if corr == 'resp':
                    corr2 = 'reso' if compare_jsons and param_to_plot == 'reso' else 'resp'
                else:
                    corr2 = 'resp' if compare_jsons and param_to_plot == 'resp' else 'reso'
                respo = resp if corr2 == 'resp' else reso 
                # print(corr, respo is reso)
                if compare_jsons and len(dims) > 2:
                    # print('hello')
                    if corr == 'resp':
                        return respo[:, i1, 0, i2]
                    else:
                        return respo[:, i1, 1, i2]
                else:
                    return respo[:, i1, i2]
            # -- response plot
            axis = ax[i2, 0]
            axis.errorbar(xval, get_resp_reso(resp, reso, corr='resp'), xerr=xerr, yerr=get_resp_reso(eresp, ereso, corr='resp'), ls='', marker='.', label=label, capsize=2)
            axis.bar(xval, 2*get_resp_reso(eresp_mc, ereso_mc, corr='resp'), bottom=get_resp_reso(resp, reso, corr='resp') - get_resp_reso(eresp_mc, ereso_mc, corr='resp'), 
                    width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            cms.polish_axis(axis, x_range=x_range, x_title=var_ref, y_title='reso.' if compare_jsons and param_to_plot == 'reso' else 'resp.', y_range=reso_range if compare_jsons and param_to_plot == 'reso' else resp_range, 
                            leg_title=leg_titles[0] if compare_jsons and len(dims)>2 else leg_title, **kwargs) 
            if param_to_plot is None or param_to_plot=='resp':
                axis.axhline(1, ls='--', color='gray')

            if hide_legend:
                axis.legend().set_visible(False) 

            # -- resolution plot
            axis = ax[i2, 1]
            axis.errorbar(xval, get_resp_reso(resp, reso, corr='reso'), xerr=xerr, yerr=get_resp_reso(eresp, ereso, corr='reso'), ls='', marker='.', label=label, capsize=2)
            axis.bar(xval, 2*get_resp_reso(eresp_mc, ereso_mc, corr='reso'), bottom=get_resp_reso(resp, reso, corr='reso') - get_resp_reso(eresp_mc, ereso_mc, corr='reso'), 
                    width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            cms.polish_axis(axis, x_range=x_range, x_title=var_ref, y_title='resp.' if compare_jsons and param_to_plot == 'resp' else 'reso.', y_range=resp_range if compare_jsons and param_to_plot == 'resp' else reso_range, 
                            leg_title=leg_titles[-1] if compare_jsons and len(dims)>2 else leg_title, **kwargs)
            if compare_jsons and param_to_plot=='resp' or jsons_mode == 'ratio':
                axis.axhline(1, ls='--', color='gray')
            if hide_legend:
                axis.legend().set_visible(False) 
            

    return figs, ax


def plot_results_from_json2G(json_file: Union[str, Path, Dict], dims: List=None, x_range=(), x_unit='', 
              cat_latex:Dict=None, mu_range=(0.78,1.02), reso2_range=(-0.001,0.10), frac_range=(0.78,1.02), **kwargs) -> Tuple[List[Figure], Axes]:
    """Create a list of Axes to plot the results in the json file.
    the dimensions to plot can be specified, additional dimension are averaged over.
    Up to 3 dimensions can be plotted.

    Args:
        json_file (Union[str, Path, Dict]): json file containing the results of the IJazZ SAS fit or Dict from the json file
        dims (List, optional): name dimensions to plot (as in the categories). Defaults to None.
        x_range (tuple, optional): range of the main variable (dimension 0). Defaults to ().
        x_unit (str, optional): usint of the main varaible. Defaults to ''.
        cat_latex (Dict, optional): dictionnary mapping var name (from categories) into a latex name. Defaults to None.
        resp_range (tuple, optional): y-range of the resp axis. Defaults to ().
        reso_range (tuple, optional): y-range if the reso axis. Defaults to ().
        leg_ncol (int, optional): number of columns in the legend. Defaults to 1.

    Returns:
        Tuple[Figure, Axes]: Figure and Axes of the plot.
    """
    # -- get the parameters
    params = json_file if isinstance(json_file, Dict) else parameters_from_json(json_file)

    if not cat_latex:
        # -- define cat_latex if not given
        cat_latex = {key: key for key in params['dims']}

    if dims is None:
        # -- keep all dims by default
        dims = [key for key in params['dims']]

    if len(dims) > 3:
        # -- keep only the 3 first dimensions
        dims = dims[:3]

    dims_to_keep = [np.where(params['dims'] == d)[0][0] for d in dims]
    dims_to_average = [id for id, d in enumerate(params['dims']) if not d in dims]
    perm = dims_to_keep + dims_to_average
    dims_to_average = np.arange(len(dims_to_keep), len(perm))

    def reshape_param(par_name):
        par = tf.transpose(params[par_name], perm=perm)
        
        epar = tf.transpose(params[f'e{par_name}'], perm=perm)
        epar_mc = tf.transpose(params[f'e{par_name}_mc'], perm=perm)

        # par = tf.reduce_sum((par/ epar**2), axis=dims_to_average) / tf.reduce_sum((1 / epar**2), axis=dims_to_average)
        # epar = tf.sqrt(1./ tf.reduce_sum((1 / epar**2), axis=dims_to_average))
        # epar_mc = tf.sqrt(1./ tf.reduce_sum((1 / epar_mc**2), axis=dims_to_average))

        shape = list(tf.shape(par).numpy()) + [1] * (3-tf.rank(par).numpy())
        return tf.reshape(par, shape), tf.reshape(epar, shape), tf.reshape(epar_mc, shape)

    mu, emu, emu_mc = reshape_param('mu')
    reso2, ereso2, ereso2_mc = reshape_param('reso2')
    frac, efrac, efrac_mc = reshape_param('frac')

    if emu.mean() == 1.0:
        emu -= 1.0
    if emu_mc.mean() == 1.0:
        emu_mc -= 1.0
    if ereso2.mean() == 1.0:
        ereso2 -= 1.0
    if ereso2_mc.mean() == 1.0:
        ereso2_mc -= 1.0
    if efrac.mean() == 1.0:
        efrac -= 1.0
    if efrac_mc.mean() == 1.0:
        efrac_mc -= 1.0


    # -- dim 0
    xbin = params['bins'][dims_to_keep[0]]
    xbin = tf.where(xbin == +np.inf, x_range[1] if len(x_range) > 1 else xbin[-2]*1.5, xbin)
    xbin = tf.where(xbin == -np.inf, x_range[0] if len(x_range) > 0 else xbin[1]/1.5, xbin)
    xval = 0.5*(xbin[1:] + xbin[:-1])
    xerr = 0.5*(xbin[1:] - xbin[:-1])
    var_ref = cat_latex[dims[0]] + x_unit

    # -- higher dimensions
    var1, var2 = None, None
    bin1, bin2 = [], []
    n_dim2 = 1
    if len(dims) > 1:
        var1 = cat_latex[dims[1]] 
        bin1 = params['bins'][dims_to_keep[1]]

    if len(dims) > 2:
        var2 = cat_latex[dims[2]] 
        bin2 = params['bins'][dims_to_keep[2]]
        n_dim2 = len(bin2) - 1

    # -- only maximum of 2 line per figure
    figs = []
    axes = []
    for ifig in range(n_dim2 // 2):
        fig, ax = plt.subplots(2, 3, figsize=(18, 4.5*2), squeeze=False)
        figs.append(fig)
        axes.append(ax)
    for ifig in range(n_dim2 % 2):
        fig, ax = plt.subplots(1, 3, figsize=(18, 4.5), squeeze=False)
        figs.append(fig)
        axes.append(ax)
    ax = np.concatenate(axes, axis=0)

    for i1 in range(mu.shape[-2]):
        for i2 in range(mu.shape[-1]):
            label = "fit" if not var1 else f"${bin1[i1]:>5.4g} \leq${var1}$< {bin1[i1+1]:<5.4g}$"
            leg_title = '' if not var2 else f"${bin2[i2]:>5.4g} \leq${var2}$< {bin2[i2+1]:<5.4g}$"

            # -- mu plot
            axis = ax[i2, 0]
            axis.errorbar(xval, mu[:, i1 ,i2], xerr=xerr, yerr=emu[:, i1, i2], ls='', marker='.', label=label, capsize=2)
            axis.bar(xval, 2*emu_mc[:, i1, i2], bottom=mu[:, i1, i2] - emu_mc[:, i1, i2], 
                    width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            cms.polish_axis(axis, x_range=x_range, x_title=var_ref, y_title='mu', y_range=mu_range, 
                            leg_title=leg_title, **kwargs) 
            axis.axhline(1, ls='--', color='gray')

            # -- resolution plot
            axis = ax[i2, 1]
            axis.errorbar(xval, reso2[:, i1 ,i2], xerr=xerr, yerr=ereso2[:, i1, i2], ls='', marker='.', label=label, capsize=2)
            axis.bar(xval, 2*ereso2_mc[:, i1, i2], bottom=reso2[:, i1, i2] - ereso2_mc[:, i1, i2], 
                    width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            cms.polish_axis(axis, x_range=x_range, x_title=var_ref, y_title='reso2.', y_range=reso2_range, 
                            leg_title=leg_title, **kwargs) 
            
            # -- frac plot
            axis = ax[i2, 2]
            axis.errorbar(xval, frac[:, i1 ,i2], xerr=xerr, yerr=efrac[:, i1, i2], ls='', marker='.', label=label, capsize=2)
            axis.bar(xval, 2*efrac_mc[:, i1, i2], bottom=frac[:, i1, i2] - efrac_mc[:, i1, i2], 
                    width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            cms.polish_axis(axis, x_range=x_range, x_title=var_ref, y_title='frac.', y_range=frac_range, 
                            leg_title=leg_title, **kwargs) 

    return figs, ax


def plot_syst_from_jsons(nominal:Union[Dict, str, Path], syst_jsons:List[Union[Dict, str, Path]], fit_2g:Union[Dict, str, Path]=None,
                         nominal_syst:Union[Dict, str, Path]=None, scale_flat_syst:float=0.0, smear_flat_syst:float=0.0, **kwargs) -> Tuple[str, List[str]]:
    """Plot the contribution of each systematic uncertainties  from the json files

    Args:
        nominal (Union[Dict, str, Path]): nominal json file
        syst_jsons (List[Union[Dict, str, Path]]): list of json files with systematic variations
        fit_2g (Union[Dict, str, Path], optional): json file with the parameters of the double gaussian fit. Defaults to None.
        scale_flat_syst (float, optional): flat scale systematic uncertainty. Defaults to 0.0.
        smear_flat_syst (float, optional): flat smear systematic uncertainty. Defaults to 0.0.
    
    Returns:
    """ 
    if not isinstance(nominal, Dict):
        nominal = parameters_from_json(nominal) 
    else:
        raise ValueError('Missing nominal json file')
    
    jsons_dict = {}
    import copy
    
    params = copy.deepcopy(nominal)
    params['resp'] = nominal['eresp']
    params['reso'] = nominal['ereso']
    params['eresp'] = np.zeros_like(nominal['eresp'])
    params['ereso'] = np.zeros_like(nominal['ereso'])
    params['eresp_mc'] = np.zeros_like(nominal['eresp'])
    params['ereso_mc'] = np.zeros_like(nominal['ereso'])

    eresp = nominal['eresp']**2
    ereso = nominal['ereso']**2

    jsons_dict['e_fitter'] = params

    params = copy.deepcopy(nominal)
    params['resp'] = nominal['eresp_mc']
    params['reso'] = nominal['ereso_mc']
    params['eresp'] = np.zeros_like(nominal['eresp'])
    params['ereso'] = np.zeros_like(nominal['ereso'])
    params['eresp_mc'] = np.zeros_like(nominal['eresp'])
    params['ereso_mc'] = np.zeros_like(nominal['ereso'])

    jsons_dict['e_fitter_mc'] = params

    flat_syst = copy.deepcopy(nominal)
    flat_syst['resp'] = np.ones_like(nominal['resp']) * scale_flat_syst
    flat_syst['reso'] = np.ones_like(nominal['reso']) * smear_flat_syst
    eresp += scale_flat_syst**2
    ereso += smear_flat_syst**2
    flat_syst['eresp'] = np.zeros_like(nominal['eresp'])
    flat_syst['ereso'] = np.zeros_like(nominal['ereso'])
    flat_syst['eresp_mc'] = np.zeros_like(nominal['eresp'])
    flat_syst['ereso_mc'] = np.zeros_like(nominal['ereso'])

    jsons_dict['flat_syst'] = flat_syst


    # return plot_results_from_json(jsons_dict, **kwargs)
    

    for syst_name, syst in syst_jsons.items():
        if not isinstance(syst, Dict):
            syst = parameters_from_json(syst)
        eresp_syst = np.abs(syst['resp']-nominal['resp'])
        ereso_syst = np.abs(syst['reso']-nominal['reso'])

        eresp += eresp_syst**2
        ereso += ereso_syst**2

        params = copy.deepcopy(nominal)
        params['resp'] = eresp_syst
        params['reso'] = ereso_syst
        params['eresp'] = np.zeros_like(nominal['eresp'])
        params['ereso'] = np.zeros_like(nominal['ereso'])
        params['eresp_mc'] = np.zeros_like(nominal['eresp'])
        params['ereso_mc'] = np.zeros_like(nominal['ereso'])

        jsons_dict[syst_name] = params


    if fit_2g:
        if not isinstance(fit_2g, Dict):
            fit_2g = parameters_from_json(fit_2g)

        r = fit_2g['resp']
        s1 = fit_2g['reso']
        s2 = fit_2g['reso2']
        mu = fit_2g['mu']
        f = fit_2g['frac']

        r2g = (f + (1-f)*mu)*r
        s2g = np.sqrt(f*s1**2 + (1-f)*(mu**2)*(s2**2) + f*(1-f)*((mu-1)**2)) * r

        eresp_2g = np.abs(nominal['resp'] - r2g)
        ereso_2g = np.abs(nominal['reso'] - s2g)

        eresp += (eresp_2g)**2
        ereso += (ereso_2g)**2

        params = copy.deepcopy(nominal)
        params['resp'] = eresp_2g
        params['reso'] = ereso_2g
        params['eresp'] = np.zeros_like(nominal['eresp'])
        params['ereso'] = np.zeros_like(nominal['ereso'])
        params['eresp_mc'] = np.zeros_like(nominal['eresp'])
        params['ereso_mc'] = np.zeros_like(nominal['ereso'])

        jsons_dict['2G'] = params

    

    params = copy.deepcopy(nominal)
    params['resp'] = np.sqrt(eresp)
    params['reso'] = np.sqrt(ereso)
    params['eresp'] = np.zeros_like(nominal['eresp'])
    params['ereso'] = np.zeros_like(nominal['ereso'])
    params['eresp_mc'] = np.zeros_like(nominal['eresp'])
    params['ereso_mc'] = np.zeros_like(nominal['ereso'])

    jsons_dict['total'] = params

    if nominal_syst:
        if not isinstance(nominal_syst, Dict):
            nominal_syst = parameters_from_json(nominal_syst)
        
        params = copy.deepcopy(nominal)
        params['resp'] = nominal_syst['eresp']
        params['reso'] = nominal_syst['ereso']
        params['eresp'] = np.zeros_like(nominal['eresp'])
        params['ereso'] = np.zeros_like(nominal['ereso'])
        params['eresp_mc'] = np.zeros_like(nominal['eresp'])
        params['ereso_mc'] = np.zeros_like(nominal['ereso'])

        jsons_dict['check'] = params

    return plot_results_from_json(jsons_dict, **kwargs)

def plot_sas_fit_results(sas: IJazZSAS, cat_latex={}, x_range=None, 
                         resp_range=(0.95, 1.05), reso_range=(0, 0.05), add_injected=False,
                         use_pt_cat=False,
                         resp_range_ratio=(0.999, 1.001), reso_range_ratio=(0.80, 1.20),
                         pt_scale_std = 1.0) -> Tuple[Figure, Axes]:
    """Plot the result of the scale and smearing fit 

    Args:
        sas (IJazZSAS): object containing the fit result of the scale and smearing (as well as the data , mc samples)
        cat_latex (dict, optional): dict containing for each variable used for categorisation a latex plotting symbol (e.g. {'pt': 'p_T'}). Defaults to {}.
        x_range (tuple, optional): x axis range (can be auto). Defaults to None.
        resp_range (tuple, optional): y axis range for response. Defaults to (0.95, 1.05).
        reso_range (tuple, optional): y axis range for resolution. Defaults to (0, 0.05).
        add_injected (bool, optional): When a resolution was injected for test, plot the value (variable sas per lepton must exist filled with the injected s&s). Defaults to False.
        use_pt_cat (bool, optional): Use the pt categorisation axis (when fitting with relative pt, allows to plot back w/r to pt). Defaults to False.
        resp_range_ratio (tuple, optional): y axis range for response ratio. Defaults to (0.999, 1.001).
        reso_range_ratio (tuple, optional): y axis range for resolution ratio. Defaults to (0.80, 1.20).
        pt_scale_std (float, optional): ?. Defaults to 1.0.

    Returns:
        tuple[plt.Figure, plt.Axes]: matplotlib figure, axes with the result
    """
    # -- define cat_latex if not given
    if not cat_latex:
        cat_latex = {key: key for key in sas.categories.keys()}

    # -- up to 3 dimensions
    n_dim1 = sas.lepton_reg.shape[1] if len(sas.lepton_reg.shape) > 1 else 1
    n_dim2 = sas.lepton_reg.shape[2] if len(sas.lepton_reg.shape) > 2 else 1

    if add_injected:
        fig, ax = plt.subplots(2, 2, sharex=True, height_ratios=(4, 1), figsize=(12, 7))
    else:
        fig, ax = plt.subplots(n_dim2, 2, figsize=(12, 4.5*n_dim2))

    def tune_dimension(arr, ashape):
        if len(ashape) == 1:
            return  np.array([[arr.reshape(ashape)]]).T
        elif len(ashape) == 2:
            return  np.array([arr.reshape(ashape).T]).T
        return arr.reshape(ashape)

    lepton_cat = tune_dimension(sas.lepton_reg, sas.lepton_reg.shape)
    resp = tune_dimension(sas.resp, sas.lepton_reg.shape)
    reso = tune_dimension(sas.reso, sas.lepton_reg.shape)
    eresp = None
    ereso = None
    if sas.eresp is not None:
        eresp = tune_dimension(sas.eresp, sas.lepton_reg.shape)
        ereso = tune_dimension(sas.ereso, sas.lepton_reg.shape)
    
    eresp_mc = None
    ereso_mc = None
    if sas.eresp_mc is not None:
        eresp_mc = tune_dimension(sas.eresp_mc, sas.lepton_reg.shape)
        ereso_mc = tune_dimension(sas.ereso_mc, sas.lepton_reg.shape)

    leg_ncol = 1
    injected = None
    name_cat = sas.cfg['fitter'].get('name_cat', 'cat')
    name_pt = sas.cfg['sas'].get('name_pt_var', 'pt')
    if add_injected and 'sas1' in sas.dt.columns:
        injected = sas.dt.groupby(f'{name_cat}2')['sas2'].agg({'mean', 'std'})\
            .reindex(range(len(sas.resp)))  # -- add missing cat if necessary
        injected['std'] = injected['std'] / injected['mean']
        leg_ncol = 2


    # -- dimension 0
    categories = sas.categories.copy()
    var_ref = list(categories.keys())[0]
    bin_ref = np.array(categories.pop(var_ref), dtype=np.float32)
    if use_pt_cat and var_ref == f'r_pt':
        df_cat = pd.concat( [sas.dt[[f'{name_cat}_r_pt{ilep}', f'{name_pt}{ilep}' , f'{name_cat}{ilep}']]\
            .rename(columns={f'{name_cat}_r_pt{ilep}': f'{name_cat}_r_pt', 
                                f'{name_pt}{ilep}': name_pt, 
                                f'{name_cat}{ilep}': name_cat}) for ilep in [1, 2]])
        df_cat = df_cat.query(f'{name_cat}_r_pt >= 0').groupby(f'{name_cat}_r_pt')[name_pt].agg({'mean', 'std'})\
            .reindex(range(resp.shape[0]))  # -- add missing cat if necessary
        df_cat['std'] *= pt_scale_std  # -- for pt re-casting do not use the full std
        var_ref = f"$p_T$  (GeV)"
    else:
        df_cat = pd.concat([sas.dt[[f'{name_cat}_{var_ref}{ilep}', f'{var_ref}{ilep}']]\
            .rename(columns={f'{name_cat}_{var_ref}{ilep}': f'{name_cat}_{var_ref}', 
                            f'{var_ref}{ilep}': f'{var_ref}'}) for ilep in [1, 2]])
        df_cat = df_cat.query(f'{name_cat}_{var_ref} >= 0').groupby(f'{name_cat}_{var_ref}')[var_ref].agg({'mean', 'std'})\
            .reindex(range(resp.shape[0]))  # -- add missing cat if necessary
        var_ref = f"${cat_latex[var_ref]}$"
    xval = df_cat['mean'].values
    xerr = df_cat['std'].values
    if not x_range:
        x_range = (bin_ref[0], min(bin_ref[-1], xval[-1] + 5*xerr[-1]))  

    # -- dimension 1
    var1 = None
    if n_dim1 > 1:
        var1 = list(categories.keys())[0]
        bin1 = np.array(categories.pop(var1), dtype=np.float32)
        var1 = cat_latex[var1]

    # -- dimension 2
    var2 = None
    if n_dim2 > 1:
        var2 = list(categories.keys())[0]
        bin2 = np.array(categories.pop(var2), dtype=np.float32)
        var2 = cat_latex[var2]

    for i1 in range(n_dim1):
        for i2 in range(n_dim2):
            label = "fit" if not var1 else f"$ {bin1[i1]:>5.4g} \leq {var1} < {bin1[i1+1]:<5.4g}$"
            leg_title = '' if not var2 else f"$ {bin2[i2]:>5.4g} \leq {var2} < {bin2[i2+1]:<5.4g}$"

            # -- plot the response   
            if injected is not None:
                axis = ax[0, 0]
            else:
                axis = ax[0] if n_dim2 == 1 else ax[i2, 0] 
            axis.errorbar(xval, resp[:, i1 ,i2], xerr=xerr, yerr=eresp[:, i1, i2], ls='', marker='.', label=label, capsize=2)
            if eresp_mc is not None:
                axis.bar(xval, 2*eresp_mc[:, i1, i2], bottom=resp[:, i1, i2] - eresp_mc[:, i1, i2], 
                         width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            if injected is not None: 
                axis.plot(xval, injected.loc[lepton_cat[:, i1, i2], 'mean'], color=axis.lines[-1].get_color(), ls='--', label='injected')
                ax[1, 0].errorbar(xval, resp[:, i1 ,i2] / injected.loc[lepton_cat[:, i1, i2], 'mean'], xerr=xerr, 
                                  yerr=eresp[:, i1, i2]/ injected.loc[lepton_cat[:, i1, i2], 'mean'], ls='', marker='.', capsize=2)
                cms.polish_axis(y_title='resp.', y_range=resp_range, ax=axis, leg_title=leg_title, leg_ncol=leg_ncol)
            else:
                cms.polish_axis(x_range=x_range, x_title=var_ref, y_title='resp.', y_range=resp_range, ax=axis, leg_title=leg_title, leg_ncol=leg_ncol) 
            axis.axhline(1, ls='--', color='gray')
    
            # -- plot the resolution
            if injected is not None:
                axis = ax[0, 1]
            else:
                axis = ax[1] if n_dim2 == 1 else ax[i2, 1] 
            axis.errorbar(xval, reso[:, i1 ,i2], xerr=xerr, yerr=ereso[:, i1, i2], ls='', marker='.', label=label, capsize=2)
            if ereso_mc is not None:
                axis.bar(xval, 2*ereso_mc[:, i1, i2], bottom=reso[:, i1, i2] - ereso_mc[:, i1, i2], 
                         width=1.5*xerr, alpha=0.7, color=axis.lines[-1].get_color())
            if injected is not None: 
                axis.plot(xval, injected.loc[lepton_cat[:, i1, i2], 'std'], color=axis.lines[-1].get_color(), ls='--', label='injected')
                ax[1, 1].errorbar(xval, reso[:, i1 ,i2] / injected.loc[lepton_cat[:, i1, i2], 'std'], xerr=xerr, 
                                  yerr=ereso[:, i1, i2]/ injected.loc[lepton_cat[:, i1, i2], 'std'], ls='', marker='.', capsize=2)
            if injected is not None:
                cms.polish_axis(y_title='reso.', y_range=reso_range, ax=axis, leg_title=leg_title, leg_ncol=leg_ncol)
                cms.polish_axis(x_title=var_ref, x_range=x_range, y_title='fit / inj.', y_range=resp_range_ratio, ax=ax[1, 0])
                cms.polish_axis(x_title=var_ref, x_range=x_range, y_title='fit / inj.', y_range=reso_range_ratio, ax=ax[1, 1])
                ax[1, 0].axhline(1, ls='-', color='black')
                ax[1, 1].axhline(1, ls='-', color='black')
            else:
                cms.polish_axis(x_range=x_range, x_title=var_ref, y_title='reso.', 
                                y_range=reso_range, ax=axis, leg_title=leg_title, leg_ncol=leg_ncol)
    return fig, ax
  




import importlib.resources as pkg_resources

ijazz_path = pkg_resources.files(__name__)
ijazz_cfg = pkg_resources.files(__name__) / '../config/'

# from .dtypes import floatzz, uintzz
# from .RegionalFitter import RegionalFitter
# from .ScaleAndSmearing import IJazZSAS, compute_sas
# from .sas_utils import parameters_from_json, parameters_to_json, ijazz_shape
# from .plotting import plot_results_from_json



__doc__ = """This Modules computes the alpha matrix which allows to smear with a gaussion function an histogram
The input histogram is of dimension d_mc (MC reference) and the smeared 
one of dimension d_dt (smeared to the dataspace).
Author: fabrice.couderc@cea.fr"""

import tensorflow as tf
import numpy as np
from ijazz.dtypes import floatzz


def alpha_evt(d_min, d_max, bin_m, r_ll, s_ll):
    """
    Compute the probability migration from due to a Gaussian over-smearing.
    Per-event variation of the alpha matrix (dim(d_min) = dim(d_max) = dim(rll) = dim(sll)).
    Note the computation is done with a normalisation to the fitting region (win_z)

    Args:
        - d_min: minimum value of the mass point for data (we return a bined probability)
        - d_max: maximum value of the mass point for data (we return a bined probability)
        - bin_m: MC binning in Mee
        - r_ll: Gaussian mean (relative, gaussian mean will be mu x r_ll)
        - s_ll: Gaussian resolution(relative as well)
    Returns a 2D tensor (dim(bining_mc), dim(r_ll))
    dim(r_ll) should be the number of events or categories

    NB: this can be used to compute the integral of pi !
    """
    d_min = tf.cast(d_min, floatzz)
    d_max = tf.cast(d_max, floatzz)
    sqrt2 = tf.cast(tf.sqrt(2.0), floatzz)
    r_loc = tf.cast(tf.reshape(r_ll, (+1, -1)), floatzz)
    s_loc = tf.cast(tf.reshape(s_ll, (+1, -1)), floatzz)

    mee = tf.reshape(0.5 * (bin_m[1:] + bin_m[:-1]), (-1, +1))  # reshape for a 3D tensor
    a_ij = tf.math.erf((tf.reshape(d_max, (+1, -1))/r_loc - mee) / (sqrt2 * mee * s_loc)) - \
           tf.math.erf((tf.reshape(d_min, (+1, -1))/r_loc - mee) / (sqrt2 * mee * s_loc))
    return 0.5 * a_ij


def alpha_2d(bin_d, bin_m, r_ll, s_ll):
    """
    Compute the probability migration from due to a Gaussian over-smearing.
    Note the computation is done with a normalisation to the fitting region

    Args:
        - bin_d: data binning in Mee
        - bin_m: MC binning in Mee
        - r_ll: Gaussian mean (relative, gaussian mean, scalar number)
        - s_ll: Gaussian resolution(relative as well, , scalar number)
    Returns a 2D tensor (dim(bining_data), dim(bining_mc))
    """
    return tf.squeeze(alpha_3d(np.array([bin_d]).T, np.array([bin_m]).T, r_ll=[r_ll], s_ll=[s_ll]), [2])


def alpha_3d(b_ic, b_jc, r_ll, s_ll):
    """
    Compute the probability migration from due to a Gaussian over-smearing.
    Note the computation is done with a normalisation to the fitting region

    Args:
        - b_ic: data binning in Mee (2D)
        - b_jc: MC binning in Mee (2D)
        - r_ll: Gaussian mean (relative, gaussian mean will be mu x r_ll)
        - s_ll: Gaussian resolution(relative as well)
    Returns a 3D tensor (dim(bining_data), dim(bining_mc), dim(r_ll))
    dim(r_ll) should be the number of events or categories
    """
    sqrt2 = tf.sqrt(2.)

    r_loc = tf.expand_dims(tf.expand_dims(r_ll, axis=0), axis=0)
    s_loc = tf.expand_dims(tf.expand_dims(s_ll, axis=0), axis=0)

    m_ijc = tf.expand_dims(0.5 * (b_jc[1:,] + b_jc[:-1,]), axis=0)
    b_ijc = tf.expand_dims(b_ic, axis=1)

    a_ijc = tf.math.erf((b_ijc[1:]/r_loc  - m_ijc) / (sqrt2 * m_ijc * s_loc)) - \
            tf.math.erf((b_ijc[:-1]/r_loc - m_ijc) / (sqrt2 * m_ijc * s_loc))
    
    return 0.5 * a_ijc


def alpha_3d_2g(b_ic, b_jc, rll_sll1, rll_sll2, rll_sll3, rll_sll4):
    """
    Compute the probability migration from due to a Gaussian over-smearing.
    Note the computation is done with a normalisation to the fitting region
    
    Args:
        - b_ic: data binning in Mee (2D)
        - b_jc: MC binning in Mee (2D)
        - r_ll: Gaussian mean (relative, gaussian mean will be mu x r_ll)
        - s_ll: Gaussian resolution(relative as well)
    Returns a 3D tensor (dim(bining_data), dim(bining_mc), dim(r_ll))
    dim(r_ll) should be the number of events or categories
    """
    sqrt2 = tf.sqrt(2.)

    m_ijc = tf.expand_dims(0.5 * (b_jc[1:,] + b_jc[:-1,]), axis=0)
    b_ijc = tf.expand_dims(b_ic, axis=1)

    a_ijc_total = tf.zeros_like(m_ijc)

    # loop over the 4 gaussians
    for (f_ll, r_ll, s_ll) in [rll_sll1, rll_sll2, rll_sll3, rll_sll4]:
        r_loc = tf.expand_dims(tf.expand_dims(r_ll, axis=0), axis=0)
        s_loc = tf.expand_dims(tf.expand_dims(s_ll, axis=0), axis=0)

        a_ijc = 0.5 * f_ll *(tf.math.erf((b_ijc[1:]/r_loc  - m_ijc) / (sqrt2 * m_ijc * s_loc)) - \
                tf.math.erf((b_ijc[:-1]/r_loc - m_ijc) / (sqrt2 * m_ijc * s_loc)))
        a_ijc_total += a_ijc
    
    
    return a_ijc_total
